from .elements import (
    Arrow,
    ArrowCircle,
    Box,
    Cylinder,
    Ellipsoid,
    Line,
    Object,
    PointCloud,
    Polygon,
    Pose,
    Robot,
    Sphere,
    Text,
    Mesh,
)
from .layer import Layer
from .stage import Stage
from .interaction_feedback import (
    InteractionFeedback,
    InteractionFeedbackType,
    CommitResult,
)
from .client import Client



import abc

from typing import ClassVar, Optional

from armarx_memory.core import MemoryID
from armarx_memory.client import MemoryNameSystem, Reader, Writer, Commit, EntityUpdate


class SpecialClientBase(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def _get_default_core_segment_id(cls) -> MemoryID:
        pass

    @classmethod
    @abc.abstractmethod
    def _get_aron_class(cls) -> ClassVar:
        pass


class SpecialWriterBase(SpecialClientBase):
    def __init__(
        self,
        writer: Writer,
        core_segment_id: Optional[MemoryID] = None,
    ):
        self.writer = writer
        self.core_segment_id = core_segment_id or self._get_default_core_segment_id()

    @classmethod
    def from_mns(
        cls,
        mns: MemoryNameSystem,
        wait=True,
        core_segment_id: Optional[MemoryID] = None,
    ):
        if core_segment_id is None:
            core_segment_id = cls._get_default_core_segment_id()
        return cls(
            mns.wait_for_writer(core_segment_id)
            if wait
            else mns.get_writer(core_segment_id)
        )

    def commit(
        self,
        entity_id: MemoryID,
        time_created_usec: Optional[int] = None,
        confidence: Optional[float] = None,
        **data_kwargs,
    ):

        commit = Commit()
        commit.add(
            self.make_update(
                entity_id=entity_id,
                time_created_usec=time_created_usec,
                confidence=confidence,
                **data_kwargs,
            )
        )
        return self.writer.commit(commit)

    def make_update(
        self,
        provider_name: Optional[str] = None,
        entity_name: Optional[str] = None,
        entity_id: Optional[MemoryID] = None,
        time_created_usec: Optional[int] = None,
        confidence: Optional[float] = None,
        data=None,
        **data_kwargs,
    ) -> EntityUpdate:

        if entity_id is None:
            entity_id = self.core_segment_id.with_provider_segment_name(
                provider_name
            ).with_entity_name(entity_name)

        if data is None:
            aron_class = self._get_aron_class()
            data = [aron_class(**data_kwargs).to_aron()]
        if not isinstance(data, list):
            data = [data]
        for i in range(len(data)):
            try:
                data[i] = data[i].to_aron()
            except AttributeError:
                pass

        return EntityUpdate(
            entity_id=entity_id,
            time_created_usec=time_created_usec,
            confidence=confidence,
            instances_data=data,
        )


class SpecialReaderBase(SpecialClientBase):
    def __init__(
        self,
        reader: Reader,
        core_segment_id: Optional[MemoryID] = None,
    ):
        self.reader = reader
        self.core_segment_id = core_segment_id or self._get_default_core_segment_id()

    @classmethod
    def from_mns(
        cls,
        mns: MemoryNameSystem,
        wait=True,
        core_segment_id: Optional[MemoryID] = None,
    ):
        if core_segment_id is None:
            core_segment_id = cls._get_default_core_segment_id()
        return cls(
            mns.wait_for_reader(core_segment_id)
            if wait
            else mns.get_reader(core_segment_id)
        )



from armarx import HandUnitInterfacePrx


class Bimanual:
    @property
    def left_hand(self):
        return HandUnitInterfacePrx.get_proxy("LeftHandUnit")

    @property
    def right_hand(self):
        return HandUnitInterfacePrx.get_proxy("RightHandUnit")

    def open_hand(self, hand_name="left, right", shape_name=None):
        """
        Opens a hand or both hands

        :param hand_name: the name of the hand
        :param shape_name: the name of the hand shape
        """
        shape_name = shape_name or "Open"
        if "left" in hand_name:
            self.left_hand.setShape(shape_name)
        if "right" in hand_name:
            self.right_hand.setShape(shape_name)
        if "both" in hand_name:
            self.left_hand.setShape(shape_name)
            self.right_hand.setShape(shape_name)

    def close_hand(self, hand_name="left, right", shape_name=None):
        """
        Closes a hand or both hands

        :param hand_name: the name of the hand
        :param shape_name: the name of the hand shape
        """
        shape_name = shape_name or "Close"
        if "left" in hand_name:
            self.left_hand.setShape(shape_name)
        if "right" in hand_name:
            self.right_hand.setShape(shape_name)
        if "both" in hand_name:
            self.left_hand.setShape(shape_name)
            self.right_hand.setShape(shape_name)



"""
This module provides functionality for receiving and providing point clouds in ArmarX.

Classes:
- PointCloudProvider: Can provide point clouds as numpy arrays.
"""

from typing import Tuple, List

import numpy as np

from visionx import PointCloudProviderInterfacePrx
from visionx import PointCloudProviderInterface
from visionx import PointCloudProcessorInterfacePrx
from visionx import PointCloudProcessorInterface
from visionx import MetaPointCloudFormat
from visionx import PointContentType


# Structured data types for point types defined in VisionX
# These are binary compatible with the Blob data used by PointCloudProvider
dtype_point_xyz = np.dtype([("position", np.float32, (3,))])
dtype_point_color_xyz = np.dtype([("color", np.uint32), ("position", np.float32, (3,))])
dtype_point_normal_xyz = np.dtype(
    [("normal", np.float32, (3,)), ("position", np.float32, (3,))]
)
dtype_point_color_normal_xyz = np.dtype(
    [("color", np.uint32), ("normal", np.float32, (3,)), ("position", np.float32, (3,))]
)
dtype_point_xyz_label = np.dtype([("position", np.float32, (3,)), ("label", np.int32)])
dtype_point_xyz_color_label = np.dtype(
    [("position", np.float32, (3,)), ("color", np.uint32), ("label", np.int32)]
)
dtype_point_xyz_intensity = np.dtype(
    [("position", np.float32, (3,)), ("intensity", np.float32)]
)


# Color as RGBA
dtype_point_rgba_xyz = np.dtype(
    [
        ("r", np.uint8),
        ("g", np.uint8),
        ("b", np.uint8),
        ("a", np.uint8),
        ("position", np.float32, (3,)),
    ]
)
dtype_point_rgba_normal_xyz = np.dtype(
    [
        ("r", np.uint8),
        ("g", np.uint8),
        ("b", np.uint8),
        ("a", np.uint8),
        ("normal", np.float32, (3,)),
        ("position", np.float32, (3,)),
    ]
)
dtype_point_xyz_rgba_label = np.dtype(
    [
        ("position", np.float32, (3,)),
        ("r", np.uint8),
        ("g", np.uint8),
        ("b", np.uint8),
        ("a", np.uint8),
        ("label", np.int32),
    ]
)


dtype_color_to_rgba_dict = {
    dtype_point_color_xyz: dtype_point_rgba_xyz,
    dtype_point_color_normal_xyz: dtype_point_rgba_normal_xyz,
    dtype_point_xyz_color_label: dtype_point_xyz_rgba_label,
}
dtype_rgba_to_color_dict = {v: k for k, v in dtype_color_to_rgba_dict.items()}


def dtype_from_point_type(point_type: PointContentType):
    if point_type == PointContentType.ePoints:
        return dtype_point_xyz
    if point_type == PointContentType.eColoredPoints:
        return dtype_point_color_xyz
    if point_type == PointContentType.eOrientedPoints:
        return dtype_point_normal_xyz
    if point_type == PointContentType.eColoredOrientedPoints:
        return dtype_point_color_normal_xyz
    if point_type == PointContentType.eLabeledPoints:
        return dtype_point_xyz_label
    if point_type == PointContentType.eColoredLabeledPoints:
        return dtype_point_xyz_color_label
    if point_type == PointContentType.eIntensity:
        return dtype_point_xyz_intensity
    raise Exception("PointContentType not yet implemented!", point_type)


def point_type_from_dtype(dt: np.dtype):
    if dt == dtype_point_xyz:
        return PointContentType.ePoints
    if dt == dtype_point_color_xyz:
        return PointContentType.eColoredPoints
    if dt == dtype_point_normal_xyz:
        return PointContentType.eOrientedPoints
    if dt == dtype_point_color_normal_xyz:
        return PointContentType.eColoredOrientedPoints
    if dt == dtype_point_xyz_label:
        return PointContentType.eLabeledPoints
    if dt == dtype_point_xyz_color_label:
        return PointContentType.eColoredLabeledPoints
    if dt == dtype_point_xyz_intensity:
        return PointContentType.eIntensity
    raise Exception("Structured data type not known!", dt)


def get_point_cloud_format(max_points: int, point_dt: np.dtype) -> MetaPointCloudFormat:
    result = MetaPointCloudFormat()
    result.size = max_points * point_dt.itemsize
    result.capacity = result.size
    result.timeProvided = 0
    result.width = max_points
    result.height = 1
    result.type = point_type_from_dtype(point_dt)
    result.seq = 0
    return result


def rgb_to_uint32(r: int, g: int, b: int):
    r, g, b = [np.clip(c, 0, 255) for c in [r, g, b]]
    return r + g * 256 + b * 256 * 256


def uint32_to_rgb(color: int) -> Tuple[int, int, int]:
    r = color % 256
    color //= 256
    g = color % 256
    color //= 256
    b = color % 256
    return r, g, b


def uint32_to_rgb_array(color_array: np.ndarray) -> np.ndarray:
    rgba_dtype = dtype_color_to_rgba_dict[color_array.dtype]
    return color_array.view(rgba_dtype)


def rgb_to_uint32_array(rgba_array: np.ndarray) -> np.ndarray:
    color_dtype = dtype_rgba_to_color_dict[rgba_array.dtype]
    return rgba_array.view(color_dtype)


def crop_by_position(
    pc: np.ndarray,
    crop_min: Tuple[float, float, float],
    crop_max: Tuple[float, float, float],
) -> np.ndarray:
    from functools import reduce

    masks = [
        pc["position"][:, i] >= threshold
        for i, threshold in enumerate(crop_min)
        if threshold is not None
    ]
    masks += [
        pc["position"][:, i] <= threshold
        for i, threshold in enumerate(crop_max)
        if threshold is not None
    ]

    if len(masks) > 0:
        mask = reduce(np.logical_and, masks[1:], masks[0])
        return pc[mask]
    else:
        return pc


def make_pcd_header(
    point_cloud: np.ndarray,
    binary=True,
) -> List[str]:
    """
    Construct lines of a PCD-compatible header.
    :param point_cloud: A point cloud with structured dtype.
    :param binary: Whether point data is stored in binary (True) or ascii (False).
    :return: The lines of the header. If binary, the header is encoded to bytes.
    """

    lines = [
        "VERSION 0.7",
    ]

    fields = ["FIELDS"]
    size = ["SIZE"]
    types = ["TYPE"]
    count = ["COUNT"]

    for name, (dtype, byte_offset) in point_cloud.dtype.fields.items():
        if name == "position":
            assert dtype.shape == (
                3,
            ), f"Expect position dtype to have shape (3,), but got {dtype.shape}."
            assert (
                dtype.base == np.float32
            ), f"Expect position dtype base ot be float32, but got {dtype.base}."
            fields.append("x y z")
        elif name == "color":
            assert (
                dtype.shape == ()
            ), f"Expect color dtype to have shape (), but got {dtype.shape}."
            assert (
                dtype.base == np.uint32
            ), f"Expect position dtype base ot be uint32, but got {dtype.base}."
            fields.append("rgba")
        else:
            raise ValueError(
                f"Encountered unknown field '{name}' of dtype {dtype} in point cloud."
            )

        dim_count = dtype.shape[0] if dtype.shape else 1

        dim_size = dtype.base.itemsize
        dim_type = None
        if np.issubdtype(dtype.base, np.floating):
            dim_type = "F"
        elif np.issubdtype(dtype.base, np.integer):
            if dtype.base.name.startswith("uint"):
                dim_type = "U"
            elif dtype.base.name.startswith("int"):
                dim_type = "I"
        if dim_type is None:
            raise ValueError(
                f"Failed to interpret base {dtype.base} of dtype {dtype} as float, unsigned int or signed int."
            )

        size.append(" ".join(map(str, [dim_size] * dim_count)))
        types.append(" ".join(map(str, [dim_type] * dim_count)))
        count.append(" ".join(map(str, [1] * dim_count)))

    lines += [
        " ".join(fields),
        " ".join(size),
        " ".join(types),
        " ".join(count),
    ]

    if point_cloud.ndim == 2:
        width, height = point_cloud.shape
    else:
        width = point_cloud.size
        height = 1

    lines += [
        f"WIDTH {width}",
        f"HEIGHT {height}",
    ]

    data_format = "binary" if binary else "ascii"
    lines += [
        "VIEWPOINT 0 0 0 1 0 0 0",
        f"POINTS {width * height}",
        f"DATA {data_format}",
    ]

    # Add new lines
    lines = [f"{l}\n" for l in lines]
    if binary:
        lines = [l.encode() for l in lines]

    return lines


def store_point_cloud(
    filepath: str,
    point_cloud: np.ndarray,
):
    """
    Save the point cloud in the binary PCD format [1].

    [1] https://pcl.readthedocs.io/projects/tutorials/en/latest/pcd_file_format.html

    :param filepath: The filepath, used as-is, without adding an extension.
    :param point_cloud: The point cloud data with a structured dtype.
    """
    binary = True  # Non-binary storage is not implemented, yet.

    header_lines = make_pcd_header(point_cloud, binary=binary)

    mode = "wb" if binary else "w"
    with open(filepath, mode) as file:
        # Header
        file.writelines(header_lines)

        # Data
        if binary:
            point_cloud.flatten().view(np.ubyte).tofile(file)
        else:
            raise NotImplementedError(
                "Storing point clouds in non-binary format is not implemented."
            )


def load_point_cloud(
    filepath: str,
) -> np.ndarray:
    """
    Load a point cloud in the binary PCD format [1].

    [1] https://pcl.readthedocs.io/projects/tutorials/en/latest/pcd_file_format.html

    :param filepath: The path to the PCD file.
    :return: The point cloud with a structured dtype.
    """

    with open(filepath, "rb") as file:
        content = file.read()

    # Find end of header
    last_line_of_header = content.find("DATA".encode())
    end_of_header = content.find("\n".encode(), last_line_of_header)

    header = content[:end_of_header]
    data = content[end_of_header + 1 :]

    header_lines: List[str] = header.decode().split("\n")
    header_dict = {}
    for line in header_lines:
        key, value = line.split(" ", maxsplit=1)
        header_dict[key] = value

    known_fields = {
        "rgba": ("color", np.uint32),
        "x y z": ("position", np.float32, (3,)),
    }

    width, height = map(int, [header_dict["WIDTH"], header_dict["HEIGHT"]])
    fields: str = header_dict["FIELDS"]

    dtype_entries = []
    while len(fields) > 0:
        found = False
        for field, dtype_entry in known_fields.items():
            if fields.startswith(field):
                found = True
                # Remove entry.
                fields = fields[len(field) :].strip()
                # Handle
                dtype_entries.append(dtype_entry)

        if not found:
            raise ValueError(
                "Failed to interpret fields '" + header_dict["FIELDS"] + "'."
            )
    dtype = np.dtype(dtype_entries)

    array = np.frombuffer(data, dtype=dtype)

    if height > 1:
        array = array.reshape((width, height))

    return array



"""
This module provides remote GUI widget classes.

Widget classes:
 - Label: Display a text string.
 - LineEdit: Edit text in a box.
 - ComboBox: Select an option from a predefined list.
 - IntSpinBox: Edit an integer value in a box.
 - IntSlider: Select an integer value with a slider.
 - FloatSpinBox: Edit a floating point value in a box.
 - FloatSlider: Select a floating point value with a slider.
 - Button: Click a button.
 - CheckBox: Mark a box with a check mark.
 - ToggleButton: Toggle a button.
 - HBoxLayout: Layout widgets horizontally.
 - VBoxLayout: Layout widgets vertically.
 - GridLayout: Layout widgets in a grid.
 - GroupBox: Group widgets in box.
 - VSpacer: Add vertical spacing to fill a layout.
 - HSpacer: Add horizontal spacing to fill a layout.
"""
from typing import List, Tuple

from armarx import RemoteGuiInterfacePrx
from armarx import RemoteGui as rg
from armarx.remote_gui.ice_wrapper import make_value_variant
from armarx.remote_gui.ice_wrapper import make_widget_state
from armarx.remote_gui.ice_wrapper import TabProxy


class Widget:
    """
    Base class for widgets.

    This class provides the following properties:
     - hidden: Controls whether the widget is visible or hidden.
     - disabled: Controls whether the widget is enabled or disabled.
    """

    next_id = 0

    def __init__(self, desc: rg.Widget):
        self.desc = desc
        self.desc.name = f"Widget_{Widget.next_id}"
        Widget.next_id = Widget.next_id + 1
        self.desc.defaultValue = make_value_variant(None)
        self.desc.defaultState = make_widget_state(hidden=False, disabled=False)
        # List of rg.Widget (only Ice description)
        self.desc.children = []

        # List of Widgets (Python instances of this class)
        self.children = []

        # Tab proxy object will be set once we create the remote tab
        self.tab: TabProxy = None

    def _check_initialized(self, property_name: str):
        if self.tab is None:
            raise Exception(
                f"Cannot query widget property '{property_name}': Widget has not been initialized"
            )

    @property
    def hidden(self):
        """Returns whether the widget is currently hidden."""
        self._check_initialized("hidden")
        return self.tab.is_hidden(self.desc.name)

    @hidden.setter
    def hidden(self, new_hidden: bool):
        """Hides or shows the widget."""
        self._check_initialized("hidden")
        self.tab.set_hidden(self.desc.name, new_hidden)

    @property
    def disabled(self):
        """Returns whether the widget is currently disabled."""
        self._check_initialized("disabled")
        return self.tab.is_disabled(self.desc.name)

    @disabled.setter
    def disabled(self, new_disabled: bool):
        """Enables or disables the widget."""
        self._check_initialized("disabled")
        self.tab.set_disabled(self.desc.name, new_disabled)


class ValueWidget(Widget):
    """
    Base class for widgets, which store values (e.g. IntSpinBox, LineEdit).

    This class provides the 'value' property, whose type is determined by the derived class.
    A LineEdit will accept 'str' values while an IntSpinBox accepts 'int' values.
    """

    def __init__(self, desc: rg.Widget):
        super().__init__(desc)

    @property
    def value(self):
        """Returns the current values stored in the widget."""
        return self.tab.get_value(self.desc.name)

    @value.setter
    def value(self, new_value):
        """
        Sets the value of the widget.

        The new value will be stored locally until the updates have been sent to the remote server.
        Sending the updates can be done via :meth:`Client.send_updates` or :meth:`Tab.send_updates`.
        """
        self.desc.defaultValue = make_value_variant(new_value)

        if self.tab is not None:
            self.tab.set_value(self.desc.name, new_value)

    def has_value_changed(self) -> bool:
        """Indicates whether the value has changed since the last update."""
        return self.tab.has_value_changed(self.desc.name)


class Label(ValueWidget):
    """The label widget displays its value as a text string."""

    def __init__(self, text: str = ""):
        super().__init__(rg.Label())
        self.value = text


class LineEdit(ValueWidget):
    """A line edit displays its value as an editable text box."""

    def __init__(self):
        super().__init__(rg.LineEdit())
        self.value = ""


class ComboBox(ValueWidget):
    """
    A combo box displays its value as a text string selected from a list of
    predefined options.
    """

    def __init__(self, options: List[str] = None):
        super().__init__(rg.ComboBox())
        if options is None:
            self.desc.options = []
        elif not options:
            raise Exception("Options must not be empty", options)
        else:
            self.desc.options = options
            self.value = options[0]

    def add_option(self, option: str):
        """Add an option to the list of predefined values of this combo box."""
        self.desc.options.append(option)

        # Set the default value to the first option that is set
        if self.desc.defaultValue.type == rg.ValueVariantType.VALUE_VARIANT_EMPTY:
            self.desc.defaultValue = make_value_variant(option)

    @property
    def options(self):
        """Returns the list of predefined options of this combo box."""
        return self.desc.options

    @options.setter
    def options(self, new_options: list):
        """Sets the list of predefined options of this combo box."""
        self.desc.options = new_options

        # Set the default value to the first option that is set
        if self.desc.defaultValue.type == rg.ValueVariantType.VALUE_VARIANT_EMPTY:
            self.desc.defaultValue = make_value_variant(new_options[0])

    @property
    def index(self) -> int:
        """Returns the index of the current value in the list of predefined options."""
        value: str = self.value
        for index, option in enumerate(self.desc.options):
            if option == value:
                return index
        return -1

    @index.setter
    def index(self, new_index: int):
        """Sets the current value to the entry of predefined options with the given index."""
        if new_index < 0 or new_index >= len(self.desc.options):
            raise Exception(
                "Index out of range for combo box",
                new_index,
                "Options:",
                self.desc.options,
            )
        self.value = self.desc.options[new_index]


class IntSpinBox(ValueWidget):
    """
    An int spin box displays its value as an integer in a text box with up-down arrows for editing.

    The 'range' property defines the allowed range of values by specifying a
    minimum and maximum value.
    The current value of the widget can only be set between these two values.
    The GUI does not allow the user to change the value outside of the defined range.
    """

    def __init__(self, value=0, range_min=0, range_max=1):
        super().__init__(rg.IntSpinBox())
        self.range = (range_min, range_max)
        self.value = value

    @property
    def range(self) -> Tuple[int, int]:
        """Returns the allowed value range as tuple (minimum value, maximum value)."""
        return self.desc.min, self.desc.max

    @range.setter
    def range(self, range_tuple: Tuple[int, int]):
        """Sets the allowed value range as tuple (minimum value, maximum value)."""
        self.desc.min = int(range_tuple[0])
        self.desc.max = int(range_tuple[1])


class IntSlider(ValueWidget):
    """
    An int slider displays its value as a horizontal slider.

    The 'range' property defines the allowed range of values by specifying a
    minimum and maximum value.
    The current value of the widget can only be set between these two values.
    The GUI does not allow the user to change the value outside of the defined range.
    """

    def __init__(self, value: int = 0, range_min: int = 0, range_max: int = 1):
        super().__init__(rg.IntSlider())
        self.range = (range_min, range_max)
        self.value = value

    @property
    def range(self) -> Tuple[int, int]:
        """Returns the allowed value range as tuple (minimum value, maximum value)."""
        return self.desc.min, self.desc.max

    @range.setter
    def range(self, range_tuple: Tuple[int, int]):
        """Sets the allowed value range as tuple (minimum value, maximum value)."""
        self.desc.min = int(range_tuple[0])
        self.desc.max = int(range_tuple[1])


class FloatSpinBox(ValueWidget):
    """
    A float spin box displays its value as a decimal number in a text box with up-down arrows for editing.

    The 'range' property defines the allowed range of values by specifying a minimum and maximum value.
    The current value of the widget can only be set between these two values.
    The GUI does not allow the user to change the value outside of the defined range.

    The 'steps' property defines in how many steps the value can be changed from its minimum to its maximum.

    The 'decimals' property defines how many decimal places of the current value are displayed in the widget.
    """

    def __init__(
        self,
        value: float = 0.0,
        range_min: float = 0.0,
        range_max: float = 1.0,
        steps: int = 100,
        decimals: int = 3,
    ):
        super().__init__(rg.FloatSpinBox())
        self.value = value
        self.range = (range_min, range_max)
        self.steps = steps
        self.decimals = decimals

    @property
    def range(self) -> Tuple[float, float]:
        """Returns the allowed value range as tuple (minimum value, maximum value)."""
        return self.desc.min, self.desc.max

    @range.setter
    def range(self, range_tuple: Tuple[float, float]):
        """Sets the allowed value range as tuple (minimum value, maximum value)."""
        self.desc.min = float(range_tuple[0])
        self.desc.max = float(range_tuple[1])

    @property
    def steps(self) -> int:
        """Returns the number of steps in which the value can be changed from minimum to maximum."""
        return self.desc.steps

    @steps.setter
    def steps(self, new_steps: int):
        """Sets the number of steps in which the value can be changed from minimum to maximum."""
        self.desc.steps = int(new_steps)

    @property
    def decimals(self) -> int:
        """Returns the number of decimal places, which are displayed in the widget."""
        return self.desc.decimals

    @decimals.setter
    def decimals(self, new_decimals: int):
        """Sets the number of decimal places, which are displayed in the widget."""
        self.desc.decimals = int(new_decimals)


class FloatSlider(ValueWidget):
    """
    A float slider displays its value as a horizontal slider.

    The 'range' property defines the allowed range of values by specifying a
    minimum and maximum value.
    The current value of the widget can only be set between these two values.
    The GUI does not allow the user to change the value outside of the defined range.

    The 'steps' property defines in how many steps the value can be changed
    from its minimum to its maximum.
    """

    def __init__(
        self,
        value: float = 0.0,
        range_min: float = 0.0,
        range_max: float = 1.0,
        steps: int = 100,
    ):
        super().__init__(rg.FloatSlider())
        self.value = value
        self.range = (range_min, range_max)
        self.steps = steps

    @property
    def range(self) -> Tuple[float, float]:
        """Returns the allowed value range as tuple (minimum value, maximum value)."""
        return self.desc.min, self.desc.max

    @range.setter
    def range(self, range_tuple: Tuple[float, float]):
        """Sets the allowed value range as tuple (minimum value, maximum value)."""
        self.desc.min = float(range_tuple[0])
        self.desc.max = float(range_tuple[1])

    @property
    def steps(self) -> int:
        """Returns the number of steps in which the value can be changed from minimum to maximum."""
        return self.desc.steps

    @steps.setter
    def steps(self, new_steps: int):
        """Sets the number of steps in which the value can be changed from minimum to maximum."""
        self.desc.steps = int(new_steps)


class Button(ValueWidget):
    """
    A button can be clicked.

    Internally a button is still a value widget with type 'int'.
    The value stores the number of times a button has been clicked.

    The 'label' property defines the text displayed on the button.
    """

    def __init__(self, label: str = "Button"):
        super().__init__(rg.Button())
        self.value = 0
        self.label = label

    @property
    def label(self) -> str:
        """Return the label text."""
        return self.desc.label

    @label.setter
    def label(self, new_label: str):
        """Set the label text."""
        self.desc.label = new_label

    def was_clicked(self):
        """Return whether the button has been clicked."""
        return self.tab.was_button_clicked(self.desc.name)


class CheckBox(ValueWidget):
    """A check box displays its boolean value as a check mark in a box."""

    def __init__(self, checked: bool = False):
        super().__init__(rg.CheckBox())
        self.value = checked


class ToggleButton(ValueWidget):
    """
    A toggle button displays its boolean value as a pushed in button.

    The 'label' property defines the text displayed on the button.
    """

    def __init__(self, label: str = "Button", toggled: bool = False):
        super().__init__(rg.ToggleButton())
        self.value = toggled
        self.label = label

    @property
    def label(self) -> str:
        """Return the label text."""
        return self.desc.label

    @label.setter
    def label(self, new_label: str):
        """Set the label text."""
        self.desc.label = new_label


class ContainerWidget(Widget):
    """This is the base class for all widgets containing child widgets."""

    def __init__(self, desc: rg.Widget, children: List[Widget] = None):
        super().__init__(desc)
        if children is not None:
            self.add_children(children)

    def add_child(self, child: Widget):
        """Add a widget to the list of child widgets."""
        self.desc.children.append(child.desc)
        self.children.append(child)

    def add_children(self, children: List[Widget]):
        """Add a list of widgets to the list of child widgets."""
        for child in children:
            self.add_child(child)

    def set_children(self, children: List[Widget]):
        """Set the list of child widgets (overwrites previous children)."""
        self.desc.children = [child.desc for child in children]
        self.children = children

    def set_child(self, child: Widget):
        """Set a single child widget (overwrites previous children)."""
        self.set_children([child])


class HBoxLayout(ContainerWidget):
    """A container widget with a horizontal box layout for the child widgets."""

    def __init__(self, children: List[Widget] = None):
        super().__init__(rg.HBoxLayout(), children)


class VBoxLayout(ContainerWidget):
    """A container widget with a vertical box layout for the child widgets."""

    def __init__(self, children: List[Widget] = None):
        super().__init__(rg.VBoxLayout(), children)


class GridLayout(ContainerWidget):
    """A container widget with a grid layout for the child widgets."""

    def __init__(self):
        super().__init__(rg.GridLayout())
        self.desc.childrenLayoutInfo = []

    def add(
        self, child: Widget, pos: Tuple[int, int] = None, span: Tuple[int, int] = None
    ):
        """Add a child widget at the specified grid position (x, y) and span (width, height)."""

        if span is None:
            span = (1, 1)

        self.add_child(child)

        layout = rg.GridLayoutData()
        layout.row = pos[0]
        layout.col = pos[1]
        layout.spanRow = span[0]
        layout.spanCol = span[1]
        self.desc.childrenLayoutInfo.append(layout)

        return self


class GroupBox(ContainerWidget):
    """
    A group box is a container widget which can only hold a single child widget.

    The 'label' property is displayed at the top of the border of the group box.
    The 'collapsed' property determines whether the group box is collapsed or expanded.
    """

    def __init__(self, label: str = "GroupBox", child: Widget = None):
        if child is None:
            super().__init__(rg.GroupBox())
        else:
            super().__init__(rg.GroupBox(), [child])
        self.label = label

    @property
    def label(self) -> str:
        """Return the label text."""
        return self.desc.label

    @label.setter
    def label(self, new_label: str):
        """Set the label text."""
        self.desc.label = new_label

    @property
    def collapsed(self) -> bool:
        """Return whether the group box was initially collapsed or expanded."""
        return self.desc.collapsed

    @collapsed.setter
    def collapsed(self, new_collapsed: bool):
        """Set whether the group box is initially collapsed or expanded."""
        self.desc.collapsed = new_collapsed


class VSpacer(Widget):
    """A spacer that fills vertical space until the parent widget is filled."""

    def __init__(self):
        super().__init__(rg.VSpacer())


class HSpacer(Widget):
    """A spacer that fills horizontal space until the parent widget is filled."""

    def __init__(self):
        super().__init__(rg.HSpacer())



import math
import numpy as np

from armarx_core import ice_manager

from visionx import StereoCalibrationInterfacePrx
from visionx import ImageProviderInterfacePrx
from visionx import MonocularCalibration

from typing import Dict

import logging


logger = logging.getLogger(__name__)


def build_calibration_matrix(
    calibration: Dict[str, float], scale: float = None
) -> np.ndarray:
    """
    Converts calibration parameters stored as a dictionary to a matrix with the
    intrinsic camera parameters.

    .. highlight:: python
    .. code-block:: python

        calibration = get_stereo_calibration('RCImageProvider')
        K = build_calibration_matrix(calibration['left'])

        import cv2
        points = np.float32([[100, 0, 0]])
        t = np.float32([0, 0, 0])
        rot_rec = np.float32([1, 0, 0])
        cv2.projectPoints(points, rot_vec, t, K, (0, 0, 0, 0))
        ...

    :param calibration: camera parameters
    :param scale: if the image is scaled
    :returns: the intrinsic camera parameters as matrix
    """
    fx = calibration["fx"]
    fy = calibration["fy"]
    cx = calibration.get("cx", None)
    cy = calibration.get("cy", None)
    if not cx:
        cx = calibration["width"] / 2.0
        cy = calibration["height"] / 2.0
    K = np.array([[fx, 0, cx], [0, fy, cy], [0, 0, 1]])
    if scale:
        K = K * scale
        K[2, 2] = 1
    return K


def get_stereo_calibration(provider_name: str):
    """
    Retrieves the camera calibration of an ArmarX ImageProvider via Ice.
    Calibration parameters are returned as dictionary.

    ..see:: build_calibration_matrix() to get a intrinsic calibration matrix

    :param provider_name: name of the component to connect to
    :returns: the calibration as dict
    """
    proxy = ImageProviderInterfacePrx.get_proxy(provider_name)
    image_format = proxy.getImageFormat()
    width = image_format.dimension.width
    height = image_format.dimension.height

    proxy = StereoCalibrationInterfacePrx.get_proxy(provider_name)
    frame = proxy.getReferenceFrame()
    stereo_calibration = proxy.getStereoCalibration()

    left_fx = stereo_calibration.calibrationLeft.cameraParam.focalLength[0]
    left_fy = stereo_calibration.calibrationLeft.cameraParam.focalLength[1]

    right_fx = stereo_calibration.calibrationRight.cameraParam.focalLength[0]
    right_fy = stereo_calibration.calibrationRight.cameraParam.focalLength[1]

    left_calibration = {
        "fx": left_fx,
        "fy": left_fy,
        "width": width,
        "height": height,
        "vertical_fov": 2.0 * math.atan(height / (2.0 * left_fy)),
        "horizontal_fov": 2.0 * math.atan(width / (2.0 * left_fx)),
    }

    right_calibration = {
        "fx": right_fx,
        "fy": right_fy,
        "width": width,
        "height": height,
        "vertical_fov": 2.0 * math.atan(height / (2.0 * right_fy)),
        "horizontal_fov": 2.0 * math.atan(width / (2.0 * right_fx)),
    }

    return {"left": left_calibration, "right": right_calibration, "frame": frame}


def get_armarx_stereo_calibration(provider_name: str):
    """
    Retrieves the camera calibration of an ArmarX ImageProvider via Ice.
    Calibration parameters are returned as dictionary of 'visionx.MonocularCalibration's.

    ..see:: image_to_world_coordinates() to convert points using this calibration

    :param provider_name: name of the component to connect to
    :returns: a dictionary of 'visionx.MonocularCalibration's
    """
    proxy = ice_manager.wait_for_proxy(StereoCalibrationInterfacePrx, provider_name)
    stereo_calibration = proxy.getStereoCalibration()
    return stereo_calibration


class MonocularCalibrationUtility(object):
    """
    Assumably faster version than using get_armarx_stereo_calibration() and image_to_world_coordinate() individually,
    as it stores certain results until the next call of image_to_world_coordinates.

    The image_to_world_coordinates() method does NOT yet take distortion parameters into account!
    Aside from that, the implementation follows <ivt/src/ivt/Calibration/Calibration.cpp>

    This visionx.camera_utils.MonocularCalibrationUtility class is different from visionx.MonocularCalibration, which
    is just a data type without business methods.
    """

    def __init__(
        self, provider_name: str, image_coordinates_are_normalized: bool = False
    ):
        """
        Retrieve the stereo calibration from the provider, check that left and right calibration are equal (monocular
        camera), and use the left one.

        :param provider_name: Name of the component that provides the stereo calibration
        """
        # store parameters
        self.stereo_calibration_provider_name = provider_name
        self.image_coordiantes_are_normalized = image_coordinates_are_normalized

        # initialization with neutral values
        self.calibration = None
        self.image_size_multiplier = np.ones(2)
        self.world_T_camera = np.identity(4)

        # actual initialization, incl. ice calls
        self._init_calibration()
        self._init_image_to_world_transformation()

    def _init_calibration(self):
        logger.debug(
            'Waiting for StereoCalibrationInterfaceProxy from component with name "{}"'.format(
                self.stereo_calibration_provider_name
            )
        )
        proxy = ice_manager.wait_for_proxy(
            StereoCalibrationInterfacePrx, self.stereo_calibration_provider_name
        )
        logger.debug("Retrieved StereoCalibrationInterfaceProxy")

        stereo_calibration = proxy.getStereoCalibration()
        if (
            not stereo_calibration.calibrationLeft
            == stereo_calibration.calibrationRight
        ):
            raise ValueError(
                "left and right calibration do not match, "
                "which violates the assumption of using a single RGBD camera"
            )
        self.calibration = stereo_calibration.calibrationLeft

    def _init_image_to_world_transformation(self):
        if self.image_coordiantes_are_normalized:
            self.image_size_multiplier = (
                self.calibration.cameraParam.width,
                self.calibration.cameraParam.height,
            )
        self.world_T_camera[0:3, 3] = np.array(self.calibration.cameraParam.translation)
        self.world_T_camera[0:3, 0:3] = np.array(self.calibration.cameraParam.rotation)

    def image_to_world_coordinates(self, image_Pt_point2D: np.ndarray, zc: float):
        camera_Pt_point_hom = np.zeros(4)
        world_Pt_point_hom = np.zeros(4)
        camera_Pt_point_hom[:2] = (
            (
                image_Pt_point2D * self.image_size_multiplier
                - self.calibration.cameraParam.principalPoint
            )
            / self.calibration.cameraParam.focalLength
            * zc
        )
        camera_Pt_point_hom[2] = zc

        world_Pt_point_hom = self.world_T_camera.dot(camera_Pt_point_hom)

        return world_Pt_point_hom[:3]


def image_to_world_coordinates(
    image_Pt_point: np.ndarray,
    zc: float,
    calibration: MonocularCalibration,
    image_coordinates_are_normalized: bool = False,
    world_T_camera: np.ndarray = None,
) -> np.ndarray:
    """
    Convert a point in an image to world coordinates.

    The method does NOT yet take distortion parameters into account!
    Aside from that, the implementation follows <ivt/src/ivt/Calibration/Calibration.cpp>

    :param image_Pt_point: 2D point in image frame
    :param zc: depth of the point in camera frame ('z camera')
    :param calibration: camera parameters
    :param world_T_camera: 4x4 homogenous matrix to transform the camera frame into the world frame
    :return: point in world frame
    """
    camera_Pt_point_hom = np.zeros(4)
    world_Pt_point_hom = np.zeros(4)

    x_multiplier = (
        calibration.cameraParam.width if image_coordinates_are_normalized else 1
    )
    y_multiplier = (
        calibration.cameraParam.height if image_coordinates_are_normalized else 1
    )

    camera_Pt_point_hom[0] = (
        (image_Pt_point[0] * x_multiplier - calibration.cameraParam.principalPoint[0])
        / calibration.cameraParam.focalLength[0]
        * zc
    )
    camera_Pt_point_hom[1] = (
        (image_Pt_point[1] * y_multiplier - calibration.cameraParam.principalPoint[1])
        / calibration.cameraParam.focalLength[1]
        * zc
    )
    camera_Pt_point_hom[2] = zc

    if world_T_camera is None:
        world_T_camera = np.identity(4)
        world_T_camera[0:3, 3] = np.array(calibration.cameraParam.translation)
        world_T_camera[0:3, 0:3] = np.array(calibration.cameraParam.rotation)

    world_Pt_point_hom = world_T_camera.dot(camera_Pt_point_hom)
    world_Pt_point = world_Pt_point_hom[:3]

    return world_Pt_point



from .dataclass_from_to_aron_ice import dataclass_from_aron_ice, dataclass_to_aron_ice
from .options import ConversionOptions


def to_aron(value) -> "armarx.aron.data.dto.GenericData":
    from armarx_memory.aron.aron_dataclass import AronDataclass

    if isinstance(value, AronDataclass):
        return value.to_aron_ice()
    else:
        from .pythonic_from_to_aron_ice import pythonic_to_aron_ice

        return pythonic_to_aron_ice(value)


def from_aron(data: "armarx.aron.data.dto.GenericData"):
    from .pythonic_from_to_aron_ice import pythonic_from_aron_ice

    return pythonic_from_aron_ice(data)



import time
from typing import Dict
from functools import lru_cache

from armarx_robots.basic_robot import Robot

from armarx import KinematicUnitInterfacePrx

# from armarx import KinematicUnitObserverInterfacePrx

from armarx import ControlMode

from armarx_robots.statechart import StatechartExecutor


class A6(Robot):
    """
    ARMAR-6

    .. highlight:: python
    .. code-block:: python

        from armarx.robots import A6
        robot = A6()
        robot.say('hello world')
    """

    profile_name = "Armar6Real"

    def __init__(self):
        super().__init__()
        self.both_arms_joint_names = [
            "ArmL1_Cla1",
            "ArmL2_Sho1",
            "ArmL3_Sho2",
            "ArmL4_Sho3",
            "ArmL5_Elb1",
            "ArmL6_Elb2",
            "ArmL7_Wri1",
            "ArmL8_Wri2",
            "ArmR1_Cla1",
            "ArmR2_Sho1",
            "ArmR3_Sho2",
            "ArmR4_Sho3",
            "ArmR5_Elb1",
            "ArmR6_Elb2",
            "ArmR7_Wri1",
            "ArmR8_Wri2",
        ]
        self.on_connect()

        # self.poses =

    def on_connect(self):
        super().on_connect()
        # self.kinematic_observer = KinematicUnitObserverInterfacePrx.get_proxy('Armar6KinematicUnitObserver')

    @property
    @lru_cache(1)
    def kinematic_unit(self):
        return KinematicUnitInterfacePrx.get_proxy("Armar6KinematicUnit")

    def init_pose(self):
        """
        Sets the joint to a default pose
        """
        joint_angles = {
            "ArmL1_Cla1": 0.036781,
            "ArmL2_Sho1": 0.839879,
            "ArmL3_Sho2": 0.111953,
            "ArmL4_Sho3": 0.178885,
            "ArmL5_Elb1": 1.317399,
            "ArmL6_Elb2": -0.077956,
            "ArmL7_Wri1": 0.081407,
            "ArmL8_Wri2": 0.171840,
            "ArmR1_Cla1": -0.036818,
            "ArmR2_Sho1": -0.839400,
            "ArmR3_Sho2": 0.111619,
            "ArmR4_Sho3": -0.179005,
            "ArmR5_Elb1": 1.319545,
            "ArmR6_Elb2": 0.078254,
            "ArmR7_Wri1": -0.081144,
            "ArmR8_Wri2": 0.171887,
            "Neck_1_Yaw": 0.0,
            "Neck_2_Pitch": 0.2,
            "TorsoJoint": 0.5,
        }
        self.move_joints(joint_angles)

    def save_pose(self, pose_name: str):
        """
        ..todo:: retrieve current pose and store under name
        """
        pass

    def set_pose(self, pose_name: str):
        """ """
        pass

    def wait_for_joints(self, joint_angles: Dict[str, float], eps=0.1, timeout=5):
        """
        Waits until the robot has reached a pose

        :param eps: angle accuraccy in radiant
        :param timeout: timeout in seconds
        """
        start_time = time.time()
        while (time.time() - start_time) < timeout:
            has_reached = True
            actual_joint_angles = self.kinematic_unit.getJointAngles()
            for joint_name, expected_joint_angle in joint_angles.items():
                actual_joint_angle = actual_joint_angles[joint_name]
                if abs(expected_joint_angle - actual_joint_angle) > eps:
                    has_reached = False
                    break
            if has_reached:
                return True
            else:
                time.sleep(0.05)
        return False

    def both_arms_zero_torque(self, joint_names=None):
        """
        Sets zero torque mode for both arms
        """
        joint_names = joint_names or self.both_arms_joint_names
        control_mode = {n: ControlMode.eTorqueControl for n in joint_names}
        joint_torques = {n: 0 for n in joint_names}
        self.kinematic_unit.switchControlMode(control_mode)
        self.kinematic_unit.setJointTorques(joint_torques)

    def both_arms_zero_velocity(self, joint_names=None):
        """
        Sets zero velocity for both arms
        """
        joint_names = joint_names or self.both_arms_joint_names
        control_mode = {n: ControlMode.eVelocityControl for n in joint_names}
        joint_velocities = {n: 0 for n in joint_names}
        self.kinematic_unit.switchControlMode(control_mode)
        self.kinematic_unit.setJointVelocities(joint_velocities)

    def move_joints(self, joint_angles: Dict[str, float]):
        """
        Sets the joint position

        :param joint_angles: A map containing the joint names and positions.
        """
        control_mode = {
            k: ControlMode.ePositionControl for k, _ in joint_angles.items()
        }
        self.kinematic_unit.switchControlMode(control_mode)
        self.kinematic_unit.setJointAngles(joint_angles)

    def place_object(self, state_parameters=None):
        s = StatechartExecutor(self.profile_name, "Armar6GraspingGroup", "PlaceObject")
        return s.run(state_parameters, True)

    def grasp_object(self, state_parameters=None):
        s = StatechartExecutor(
            self.profile_name, "Armar6GraspingGroup", "GraspSingleObject"
        )
        return s.run(state_parameters, True)



import dataclasses as dc
import typing as ty


@dc.dataclass
class ConversionOptions:
    # Explicit translations from python variable names to Aron names.
    names_python_to_aron_dict: ty.Dict[str, str] = dc.field(default_factory=dict)

    # If true, convert variable names from snake_case (python) to camelCase (Aron).
    names_snake_case_to_camel_case: bool = False

    def name_python_to_aron(self, python_name: str) -> str:
        aron_name = self.names_python_to_aron_dict.get(python_name, None)

        if aron_name is None:
            if self.names_snake_case_to_camel_case:
                from .name_conversion import snake_case_to_camel_case

                aron_name = snake_case_to_camel_case(python_name)
            else:
                aron_name = python_name

        return aron_name

    def name_aron_to_python(self, aron_name: str) -> str:
        for python, aron in self.names_python_to_aron_dict.items():
            if aron == aron_name:
                return python

        if self.names_snake_case_to_camel_case:
            from .name_conversion import camel_case_to_snake_case

            python_name = camel_case_to_snake_case(aron_name)
        else:
            python_name = aron_name

        return python_name



import os
import time

import logging
from typing import Any
from typing import TypeVar

from functools import lru_cache

import Ice
from Ice import NotRegisteredException

from IceGrid import RegistryPrx
from IceGrid import ObjectExistsException

from IceStorm import TopicManagerPrx
from IceStorm import NoSuchTopic
from IceStorm import AlreadySubscribed

from .config import get_ice_config_files
from .name_helper import get_ice_default_name

logger = logging.getLogger(__name__)


T = TypeVar("T")


def register_object(
    ice_object: Ice.Object, ice_object_name: str = None
) -> Ice.ObjectPrx:
    """
    Register a local ice object under the given name

    :param ice_object: Local ice object instance
    :param ice_object_name: Name with which the object should be registered
    :return: Proxy to this object
    """
    if not isinstance(ice_object, Ice.Object):
        logger.error("ice object is not an Ice.Object")
        raise ValueError("ice_object is not an Ice.Object")
    if not ice_object_name:
        ice_object_name = ice_object.__class__.__name__
        logger.debug("Using class name %s to register the object", ice_object_name)
    adapter = freezer().communicator.createObjectAdapterWithEndpoints(
        ice_object_name, "tcp"
    )
    ice_object_id = freezer().communicator.stringToIdentity(ice_object_name)
    adapter.add(ice_object, ice_object_id)
    adapter.activate()
    proxy = adapter.createProxy(ice_object_id)
    admin = get_admin()
    try:
        logger.info("adding new object %s", ice_object_name)
        admin.addObjectWithType(proxy, proxy.ice_id())
    except ObjectExistsException:
        logger.info("updating new object %s", ice_object_name)
        admin.updateObject(proxy)
    return proxy


def get_topic(cls: T, topic_name: str = None) -> T:
    """
    Retrieve a topic proxy casted to the first parameter

    :param cls: Type of the topic
    :param topic_name: Name of the topic
    :type topic_name: str
    :return: a casted topic proxy
    """
    topic_name = topic_name or get_ice_default_name(cls)
    topic_manager = TopicManagerPrx.checkedCast(
        freezer().communicator.stringToProxy("IceStorm/TopicManager")
    )
    topic = None
    try:
        topic = topic_manager.retrieve(topic_name)
    except NoSuchTopic:
        topic = topic_manager.create(topic_name)
    logger.info("Publishing to topic %s", topic_name)
    pub = topic.getPublisher().ice_oneway()
    return cls.uncheckedCast(pub)


def using_topic(proxy, topic_name: str = None):
    """
    .. seealso:: :func:`register_object`

    :param proxy: the instance where the topic event should be called
    :param topic_name: the name of the topic to connect to
    :type topic_name: str
    """
    topic_manager = TopicManagerPrx.checkedCast(
        freezer().communicator.stringToProxy("IceStorm/TopicManager")
    )
    topic = None
    topic_name = topic_name or get_ice_default_name(proxy.__class__)
    try:
        topic = topic_manager.retrieve(topic_name)
    except NoSuchTopic:
        topic = topic_manager.create(topic_name)
    try:
        topic.subscribeAndGetPublisher(None, proxy)
    except AlreadySubscribed:
        topic.unsubscribe(proxy)
        topic.subscribeAndGetPublisher(None, proxy)
    logger.info("Subscribing to topic %s", topic_name)
    return topic


def wait_for_dependencies(proxy_names, timeout: int = 0):
    """
    waits for a dependency list

    :param proxy_names: the proxy names to wait for
    :type proxy_names: list of str

    :returns:  True if all dependencies are resolved
    :rtype: bool
    """
    start_time = time.time()
    while not freezer().communicator.isShutdown():
        if timeout and (start_time + timeout) < time.time():
            logging.exception("Timeout while waiting for proxies %s", proxy_names)
            return False
        dependencies_resolved = True
        for proxy_name in proxy_names:
            try:
                proxy = freezer().communicator.stringToProxy(proxy_name)
                proxy.ice_ping()
            except NotRegisteredException:
                dependencies_resolved = False
        if dependencies_resolved:
            return True
        else:
            time.sleep(0.1)


def wait_for_proxy(cls, proxy_name: str = None, timeout: int = 0):
    """
    waits for a proxy.

    :param cls: the class definition of an ArmarXComponent
    :param proxy_name: name of the proxy
    :param timeout: timeout in seconds to wait for the proxy. Zero means to wait forever
    :returns: the retrieved proxy
    :rtype: an instance of cls
    """
    proxy = None
    proxy_name = proxy_name or get_ice_default_name(cls)
    start_time = time.time()
    while not freezer().communicator.isShutdown() and proxy is None:
        try:
            proxy = freezer().communicator.stringToProxy(proxy_name)
            proxy_cast = cls.checkedCast(proxy)
            return proxy_cast
        except NotRegisteredException:
            proxy = None

        if timeout and (start_time + timeout) < time.time():
            logging.exception("Timeout while waiting for proxy %s", proxy_name)
            return None
        else:
            logger.debug("Waiting for proxy %s", proxy_name)
            time.sleep(0.1)


def get_proxy(cls: T, proxy_name: str = None) -> T:
    """
    Connects to a proxy.

    :param cls: the class definition of an ArmarXComponent
    :param proxy_name: name of the proxy
    :type proxy_name: str
    :returns: the retrieved proxy
    :rtype: an instance of cls
    :raises: Ice::NotRegisteredException if the proxy is not available
    """
    proxy_name = proxy_name or get_ice_default_name(cls)
    try:
        proxy = freezer().communicator.stringToProxy(proxy_name)
        return cls.checkedCast(proxy)
    except NotRegisteredException:
        logging.exception("Proxy %s does not exist", proxy_name)


def get_admin():
    return freezer().registry.createAdminSession("user", "password").getAdmin()


def is_connected(ice_node_name: str) -> bool:
    return get_admin().pingNode(ice_node_name)


def is_alive() -> bool:
    """
    checks if shutdown has been invoked on the communicator.

    :returns: true if ice grid registry is alive
    """
    return not freezer().communicator.isShutdown()


def wait_for_shutdown():
    """
    sleeps until the ice communicator receives a shutdown signal
    or the program receives a keyboard interrupt
    """
    try:
        freezer().communicator.waitForShutdown()
    except KeyboardInterrupt:
        pass


@lru_cache(maxsize=1)
def freezer():
    """ """
    return Freezer()


def test_connection():
    if not is_connected("NodeMain"):
        logger.error("Ice is not running.")
        raise Exception("Ice is not running.")


class Freezer:

    registry: RegistryPrx = None
    communicator = None

    def __init__(self):
        ice_config_files = get_ice_config_files()
        ice_communicator = Ice.initialize(
            ["--Ice.Config={}".format(",".join(ice_config_files))]
        )
        ice_registry_proxy = ice_communicator.stringToProxy("IceGrid/Registry")
        ice_registry = RegistryPrx.checkedCast(ice_registry_proxy)

        self.communicator = ice_communicator
        self.registry = ice_registry
        import atexit

        atexit.register(self.__del__)

    @property
    def admin(self):
        return self.registry.createAdminSession("user", "password").getAdmin()

    def __enter__(self):
        return self.communicator

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def shutdown(self):
        if not self.communicator:
            return
        try:
            self.communicator.destroy()
        except Exception as ex:
            logger.error("Error while shutting down ice communicator")
        finally:
            self.communicator = None

    def __del__(self):
        self.shutdown()



import numpy as np

from typing import Dict, List, Optional

from armarx_core import slice_loader

from armarx_memory.ice_conv.ice_twin import IceTwin
from armarx_memory.ice_conv.RobotAPI.PoseBase import Vector3BaseConv, QuaternionBaseConv


class Box(IceTwin):
    """
    A Python twin of the Ice DTO class `armarx.objpose.Box`
    from `objectpose/object_pose_types.ice` in RobotAPI.
    """

    def __init__(
        self,
        position: Optional[np.ndarray] = None,
        orientation: Optional[np.ndarray] = None,
        extents: Optional[np.ndarray] = None,
    ):
        import transforms3d as tf3d

        self.position = (
            np.array(position) if position is not None else np.zeros(3, dtype=float)
        )
        self.orientation = (
            np.array(orientation)
            if orientation is not None
            else tf3d.quaternions.qeye()
        )
        self.extents = (
            np.array(extents) if extents is not None else np.zeros(3, dtype=float)
        )

        self.__vec3_conv: Optional[Vector3BaseConv] = None
        self.__quat_conv: Optional[QuaternionBaseConv] = None

    @property
    def pose(self) -> np.ndarray:
        import transforms3d as tf3d

        return tf3d.affines.compose(
            T=self.position, R=tf3d.quaternions.quat2mat(self.orientation), Z=np.ones(3)
        )

    @classmethod
    def from_aabb(cls, minim, maxim) -> "Box":
        minim = np.array(minim)
        maxim = np.array(maxim)
        position = 0.5 * (minim + maxim)
        extents = maxim - minim
        return Box(position=position, extents=extents)

    @classmethod
    def from_aabb_dict(cls, aabb_dict: Dict) -> "Box":
        return cls.from_aabb(minim=aabb_dict["min"], maxim=aabb_dict["max"])

    def get_corners(self, global_pose=np.eye(4)) -> List[np.ndarray]:
        pose = global_pose @ self.pose
        pos = pose[:3, 3]
        ori = pose[:3, :3]
        points = []
        for ext, axis in zip(self.extents, ori.T):
            for s in [-1, 1]:
                points.append(pos + s * 0.5 * ext * axis)
        return points

    def to_aabb(self, global_pose=np.eye(4)):
        import transforms3d as tf3d

        points = np.array(self.get_corners(global_pose))

        pmin = np.min(points, axis=0)
        pmax = np.max(points, axis=0)
        return Box(0.5 * (pmin + pmax), tf3d.quaternions.qeye(), pmax - pmin)

    @classmethod
    def _get_ice_cls(cls):
        slice_loader.load_armarx_slice("RobotAPI", "objectpose/object_pose_types.ice")
        from armarx.objpose import Box

        return Box

    def _set_from_ice(self, dto):
        self.position = self._vec3_conv.from_ice(dto.position)
        self.orientation = self._quat_conv.from_ice(dto.orientation)
        self.extents = self._vec3_conv.from_ice(dto.extents)

    def _set_to_ice(self, dto):
        dto.position = self._vec3_conv.to_ice(self.position)
        dto.orientation = self._quat_conv.to_ice(self.orientation)
        dto.extents = self._vec3_conv.to_ice(self.extents)

    @property
    def _vec3_conv(self) -> Vector3BaseConv:
        if self.__vec3_conv is None:
            self.__vec3_conv = Vector3BaseConv()
        return self.__vec3_conv

    @property
    def _quat_conv(self) -> QuaternionBaseConv:
        if self.__quat_conv is None:
            self.__quat_conv = QuaternionBaseConv()
        return self.__quat_conv

    def __repr__(self):
        return "<{c} pos={p} ori={o} ext={e}>".format(
            c=self.__class__.__name__,
            p=self.position,
            o=self.orientation,
            e=self.extents,
        )



import numpy as np


def make_grid(
    *x_i,
) -> np.ndarray:
    """
    Given arrays N arrays
        (a_i) (i=1..M_a), (b_j) (j=1..M_b), ..., (z_k) (k = 1..M_z),
    where each array is a 1D array containing M_* points along a dimension,
    construct a grid containing all combinations (a_i, b_j, ..., z_k) in R^N
    structured in an (N+1) dimensional array of shape (M_a, M_b, ..., M_z, N).

    For example::

        a = np.arange(4)  # shape (4,)
        b = np.arange(5)  # shape (5,)
        c = np.arange(6)  # shape (6,)
        grid = make_grid(a, b, c)  # shape (4, 5, 6, 3)
        point = grid[0, 3, 5]  # shape (3,)
        print(point)  # array([0 3 5])

    :param x_i: Arrays specifying the points along each dimension.
    :return: A grid of coordinates.
    """
    return np.moveaxis(np.stack(np.meshgrid(*x_i, indexing="ij")), 0, -1)



import dataclasses as dc
import typing as ty


@dc.dataclass
class AronDataclass:
    def to_dict(self) -> ty.Dict[str, ty.Any]:
        from armarx_memory.aron.conversion.dataclass_from_to_pythonic import (
            dataclass_to_dict,
        )

        return dataclass_to_dict(self)

    def to_aron_ice(self) -> "armarx.aron.data.dto.Dict":
        from armarx_memory.aron.conversion.dataclass_from_to_aron_ice import (
            dataclass_to_aron_ice,
        )

        return dataclass_to_aron_ice(self, options=self._get_conversion_options())

    @classmethod
    def from_dict(cls, data: ty.Dict[str, ty.Any]) -> "AronDataclass":
        from armarx_memory.aron.conversion.dataclass_from_to_pythonic import (
            dataclass_from_dict,
        )

        return dataclass_from_dict(cls, data)

    @classmethod
    def from_aron_ice(cls, data: "armarx.aron.data.dto.Dict") -> "AronDataclass":
        from armarx_memory.aron.conversion.dataclass_from_to_aron_ice import (
            dataclass_from_aron_ice,
        )

        return dataclass_from_aron_ice(cls, data, options=cls._get_conversion_options())

    @classmethod
    def _get_conversion_options(cls) -> ty.Optional["ConversionOptions"]:
        return None



"""
This module provides functionality for remote GUI clients.

Classes:
 - Tab: Base class for remote tabs.
 - Client: Remote GUI client.
"""

from armarx import RemoteGuiInterfacePrx
from armarx import RemoteGui as rg

from armarx.remote_gui.ice_wrapper import TabProxy
from armarx.remote_gui.widgets import Widget

from typing import Callable, List, Optional
import abc


class Tab:
    """
    This is a base class for remote GUI tabs.

    A remote GUI tab is identified by an id string.
    The structure of the GUI is defined via a single root widget.
    The root widget is usually a container widget which contains many child widgets.

    The :meth:`create_widget_tree` method must be overridden in derived class to create the root widget and all its children.

    The :meth:`on_update` method is called by the Client after updates have been received from the GUI.
    """

    def __init__(self, id: str):
        self.id = id
        self.root_widget = None
        self.proxy: TabProxy = None

    @abc.abstractmethod
    def create_widget_tree(self) -> Widget:
        """
        Creates a widget tree and returns the root widget of the tree.

        This method MUST be overridden by a derived class.

        In a derived class, you will create your widgets here and nest them via container widgets.
        You can store the widgets, which you want to access later (usually value widgets), as member variables.
        Widgets which are only used for layout or to display constant values can be created as local variables
        and need not be stored.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def on_update(self):
        """
        Updates the internal state after an update has been received from the GUI.

        This method CAN be overridden by a derived class.

        In a derived class, you can update your internal state according to the new updates from the GUI.
        """
        pass

    def connect(self, proxy: "RemoteGuiInterfacePrx"):
        """Connect this instance with a remote GUI tab."""
        self.root_widget = self.create_widget_tree()
        proxy.createTab(self.id, self.root_widget.desc)
        self.proxy = TabProxy(proxy, self.id)
        self.proxy.initialize_widget(self.root_widget)

    def remove(self):
        """Remove this remote GUI tab."""
        self.proxy.remove()

    def receive_updates(self):
        """Receive updates from the remote GUI."""
        self.proxy.receive_updates()

    def send_updates(self):
        """Send updates to the remote GUI."""
        self.proxy.send_updates()


class Client:
    """The client connects to the remote GUI and manages tabs."""

    def __init__(self, provider_name: str = "RemoteGuiProvider"):
        self.provider_name = provider_name
        self.proxy = RemoteGuiInterfacePrx.get_proxy(provider_name)
        self.tabs: List[Tab] = []

    def add_tab(self, tab: Tab):
        """Add a tab and establish a connection to the remote GUI."""
        tab.connect(self.proxy)
        self.tabs.append(tab)

    def remove_tab(self, tab: Tab):
        """Remove a tab from the remote GUI."""
        tab.remove()
        self.tabs.remove(tab)

    def receive_updates(self):
        """Receive updates from the remote GUI for all added tabs."""
        for tab in self.tabs:
            tab.receive_updates()
            tab.on_update()

    def send_updates(self):
        """Send updates to the remote GUI for all added tabs."""
        for tab in self.tabs:
            tab.send_updates()

    def update_loop(
        self, callback: Callable, block=True
    ) -> Optional["threading.Thread"]:
        """
        Run a loop receiving and sending updates.
        :param callback: The callback to call after receiving updates.
        :param block:

            If true, blocks until a KeyboardInterrupt is received.
            If false, starts the loop in a thread and returns the thread.

        :return: If block is False, returns the thread, otherwise returns nothing.
        """
        if block:
            try:
                while True:
                    try:
                        self.receive_updates()

                        callback()

                        self.send_updates()
                    except KeyError:
                        pass

            except KeyboardInterrupt:
                pass

        else:
            from threading import Thread

            thread = Thread(
                target=lambda: self.update_loop(callback=callback, block=True),
                name="Remote GUI",
            )
            thread.start()
            return thread



import numpy as np

from typing import List, Tuple, Union

from armarx import remote_gui as rg


class NdArrayWidget:
    """
    A grid of float or int spin boxes parametrizing a 1D or 2D numpy array.
    """

    SpinBox = Union[rg.FloatSpinBox, rg.IntSpinBox]

    def __init__(
        self,
        array: np.ndarray,
        column_vector=False,
        float_widget_cls=rg.FloatSpinBox,
        int_widget_cls=rg.IntSpinBox,
        **kwargs,
    ):
        array = np.array(array)

        if "row_vector" in kwargs:
            print(
                "NdArrayWidget(): The argument 'row_vector' was replaced by 'column_vector'.\n"
                "If you are using 'row_vector=True', use 'column_vector=False' instead, and vice-versa."
            )
            column_vector = not kwargs.pop("row_vector")

        if "int" in array.dtype.name:
            self.spin_box_type = int_widget_cls
            self.scalar_type = int
        else:
            self.spin_box_type = float_widget_cls
            self.scalar_type = float

        self.column_vector = column_vector
        self.spin_boxes = np.zeros(array.shape, dtype=np.object)
        for i in range(self.spin_boxes.size):
            sb = self.spin_box_type(value=self.scalar_type(array.flat[i]), **kwargs)
            self.spin_boxes.flat[i] = sb

    def create_tree(
        self,
    ):
        layout = rg.GridLayout()
        if self.spin_boxes.ndim == 1:
            for i, sb in enumerate(self.spin_boxes):
                layout.add(sb, (i, 0) if self.column_vector else (0, i))

        elif self.spin_boxes.ndim == 2:
            for i, row in enumerate(self.spin_boxes):
                for j, sb in enumerate(row):
                    layout.add(sb, (i, j))
        else:
            raise ValueError(
                f"{self.__class__.__name__} supports only 1- and 2-dimensional arrays, "
                f"but requested ndim is {self.spin_boxes.ndim}."
            )
        return layout

    @property
    def value(self) -> np.ndarray:
        array = np.zeros(self.spin_boxes.shape, dtype=self.scalar_type)
        for i in range(self.spin_boxes.size):
            array.flat[i] = self.spin_boxes.flat[i].value
        return array

    @value.setter
    def value(self, values: np.ndarray):
        assert (
            values.shape == self.spin_boxes.shape
        ), f"Shape mismatch (expected {self.spin_boxes.shape}, but got {values.shape})."

        for sb, val in zip(self.spin_boxes.flat, values.flat):
            sb.value = self.scalar_type(val)

    def get_array(
        self,
    ) -> np.ndarray:
        return self.value

    def has_any_changed(self):
        for sb in self.spin_boxes.flat:
            if sb.has_value_changed():
                return True
        return False

    def find_changed(self) -> List[Tuple]:
        indices = np.indices(self.spin_boxes.shape).T.reshape(-1, self.spin_boxes.ndim)
        return [
            tuple(i) for i in indices if self.spin_boxes[tuple(i)].has_value_changed()
        ]



from typing import List, Optional

from armarx_memory.aron.aron_ice_types import AronIceTypes
from armarx_memory.core import MemoryID, time_usec, DateTimeIceConverter
from armarx_memory.ice_conv import ice_twin


from armarx_core import slice_loader

slice_loader.load_armarx_slice("RobotAPI", "armem/server/MemoryInterface.ice")

import armarx.armem as armem

date_time_conv = DateTimeIceConverter()


class EntityUpdate(ice_twin.IceTwin):
    def __init__(
        self,
        entity_id: MemoryID = None,
        instances_data: List[AronIceTypes.Data] = None,
        time_created_usec: Optional[int] = None,
        confidence: Optional[float] = None,
        time_sent_usec: Optional[int] = None,
    ):

        self.entity_id: MemoryID = entity_id or MemoryID()
        self.instances_data = instances_data or []

        self.time_created_usec = time_created_usec or time_usec()

        self.confidence = 1.0 if confidence is None else float(confidence)
        self.time_sent_usec: Optional[int] = time_sent_usec

    def set_time_created_to_now(self):
        self.time_created_usec = self.now()

    @staticmethod
    def now():
        return time_usec()

    @classmethod
    def _get_ice_cls(cls):
        return armem.data.EntityUpdate

    def _set_to_ice(self, dto: armem.data.Commit):
        dto.entityID = self.entity_id.to_ice()
        dto.instancesData = self.instances_data
        dto.timeCreated = date_time_conv.to_ice(self.time_created_usec)

        dto.confidence = self.confidence
        dto.timeSent = date_time_conv.to_ice(
            self.time_sent_usec if self.time_sent_usec is not None else -1
        )

    def _set_from_ice(self, dto):
        self.entity_id.set_from_ice(dto.entityID)
        self.instances_data = dto.instancesData
        self.time_created_usec = date_time_conv.from_ice(dto.timeCreated)

        self.confidence = dto.confidence
        self.time_sent_usec = date_time_conv.from_ice(dto.timeSent)

        assert (
            self.time_created_usec > 0
        ), f"The time sent must be valid: {self.time_created_usec}"
        assert (
            self.time_sent_usec > 0 or self.time_sent_usec is None
        ), f"The time sent must be valid: {self.time_created_usec}"

    def __repr__(self):
        return "<{c} id={i} t={t} c={con:.3f} with {n} instance{ns}>".format(
            c=self.__class__.__name__,
            i=self.entity_id,
            t=self.time_created_usec,
            n=len(self.instances_data),
            ns="" if len(self.instances_data) == 1 else "s",
            con=self.confidence,
            # d=self.instances_data,
        )


class Commit(ice_twin.IceTwin):
    def __init__(self, updates: List[EntityUpdate] = None):
        self.updates = updates or []

    def add(self, update: Optional[EntityUpdate] = None, **kwargs):
        if update is None:
            update = EntityUpdate(**kwargs)
        self.updates.append(update)
        return update

    @classmethod
    def _get_ice_cls(cls):
        return armem.data.Commit

    def _set_to_ice(self, dto: armem.data.Commit):
        dto.updates = ice_twin.to_ice(self.updates)

    def _set_from_ice(self, dto):
        self.updates = EntityUpdate.from_ice(dto.updates)

    def __repr__(self):
        return "<{c} with {n} update{ns}>".format(
            c=self.__class__.__name__,
            n=len(self.updates),
            ns="" if len(self.updates) == 1 else "s",
        )

    def __str__(self):
        return "{r}\n- {u}".format(
            r=self.__repr__(), u="n- ".join(map(str, self.updates))
        )



import os
import configparser

from enum import Enum
from typing import List

import logging


logger = logging.getLogger(__name__)


def _load_config() -> configparser.ConfigParser:
    """
    :raises FileNotFoundError: if armarx.ini does not exists
    :returns: the default armarx config
    :rtype: configparser.ConfigParser
    """
    config_file = os.path.join(_get_config_dir(), "armarx.ini")
    if not os.path.exists(config_file):
        raise FileNotFoundError("ArmarX config file does not exists.")
    config_parser = configparser.ConfigParser()
    config_parser.read(config_file)
    return config_parser


def _get_config_dir():
    """
    Returns the current armarx config directory. Therefore this methods first
    checks if the configuration is given by an environment variable, or a
    workspace is currently loaded. Otherwise it tests the default path

    :returns: the current config directory
    """
    if os.environ.get("ARMARX_USER_CONFIG_DIR", None):
        config_dir = os.path.expanduser(os.environ["ARMARX_USER_CONFIG_DIR"])
    elif os.environ.get("ARMARX_WORKSPACE", None):
        config_dir = os.path.join(os.environ["ARMARX_WORKSPACE"], "armarx_config")
    else:
        config_dir = os.path.expanduser("~/.armarx/")

    if config_dir:
        config_dir = os.path.expandvars(config_dir)
    if os.path.isdir(config_dir):
        return config_dir
    else:
        raise FileNotFoundError(
            f'ArmarX config folder does not exists. (Tried: "{config_dir}")'
        )


def get_packages() -> str:
    """
    Lists all packages that are considered by the statecharts
    """
    default_packages = (
        "ArmarXCore,ArmarXGui,RobotAPI,VisionX,RobotSkillTemplates,ActiveVision"
    )
    return config.get("AutoCompletion", "packages", fallback=default_packages)


def get_ice_config_files() -> List[str]:
    """
    The default Ice.Config
    """
    config_dir = _get_config_dir()
    return [
        os.path.join(config_dir, "default.generated.cfg"),
        os.path.join(config_dir, "default.cfg"),
    ]


def get_python_build_dir() -> str:
    if os.environ.get("ARMARX_PYTHON_BUILD_DIR", None):
        return os.path.expandvars(os.environ["ARMARX_PYTHON_BUILD_DIR"])


class CodeGenerationType(Enum):
    DYNAMIC = 0
    STATIC = 1


def get_code_generation_type():
    if config.get("Misc", "PythonStaticCode", fallback=None):
        python_binary_dir = get_python_build_dir()
        if os.path.exists(python_binary_dir):
            return CodeGenerationType.STATIC
        else:
            logger.warning(
                "Static python code selected but directory %s does not exists",
                python_binary_dir,
            )
    return CodeGenerationType.DYNAMIC


config = _load_config()



from typing import List
from typing import Optional

from armarx_memory.core import MemoryID

from armarx_memory.segments.base_segment import BaseClient
from armarx_memory.segments.base_segment import GenericDataClass
from armarx_memory.segments.base_segment import BaseReader
from armarx_memory.segments.base_segment import BaseWriter


class ObjectClassClientBase(BaseClient):

    core_segment_id = MemoryID("Object", "Class").with_provider_segment_name(
        "PriorKnowledgeData"
    )

    default_entity_name = "object_class"


class ObjectClassReader(BaseReader, ObjectClassClientBase):
    pass



import enum
import dataclasses as dc
import typing as ty

from armarx_memory.aron.aron_dataclass import AronDataclass
from armarx_memory.aron.common.names import Names
from armarx_memory.aron.common.package_path import PackagePath
from armarx_memory.client import MemoryNameSystem, Commit, Reader, Writer
from armarx_memory.core import MemoryID


class Gender(enum.IntEnum):
    DIVERSE = 0
    FEMALE = 1
    MALE = 2


class Handedness(enum.IntEnum):
    BOTH = 0
    LEFT = 1
    RIGHT = 2


@dc.dataclass
class Profile(AronDataclass):

    first_name: str = ""
    last_name: str = ""

    names: Names = Names()

    gender: Gender = Gender.DIVERSE
    handedness: Handedness = Handedness.BOTH

    face_image_paths: ty.List[PackagePath] = dc.field(default_factory=list)
    birthday: int = -1

    height_in_mm: int = -1
    weight_in_g: int = -1

    favourite_color: ty.Tuple[int, int, int] = (192, 192, 192)

    roles: ty.List[str] = dc.field(default_factory=list)

    attributes: ty.Dict[str, ty.Any] = dc.field(default_factory=dict)

    @classmethod
    def _get_conversion_options(cls):
        from armarx_memory.aron.conversion import ConversionOptions

        return ConversionOptions(
            names_snake_case_to_camel_case=True,
            names_python_to_aron_dict={
                "height_in_mm": "heightInMillimeters",
                "weight_in_g": "weightInGrams",
            },
        )


class ProfileClientBase:

    core_segment_id = MemoryID("Human", "Profile")

    def __init__(self):
        pass

    def make_entity_name(self, provider_name: str, entity_name: str):
        return self.core_segment_id.with_provider_segment_name(
            provider_name
        ).with_entity_name(entity_name)


class ProfileWriter(ProfileClientBase):
    def __init__(self, writer: Writer):
        super().__init__()
        self.writer = writer

    @classmethod
    def from_mns(cls, mns: MemoryNameSystem, wait=True) -> "ProfileWriter":
        return cls(
            mns.wait_for_writer(cls.core_segment_id)
            if wait
            else mns.get_writer(cls.core_segment_id)
        )

    def commit(
        self, entity_id: MemoryID, profile: Profile, time_created_usec=None, **kwargs
    ):
        commit = Commit()
        commit.add(
            entity_id=entity_id,
            time_created_usec=time_created_usec,
            instances_data=[profile.to_aron_ice()],
            **kwargs,
        )
        return self.writer.commit(commit)


class ProfileReader(ProfileClientBase):
    def __init__(self, reader: Reader):
        super().__init__()
        self.reader = reader

    @classmethod
    def from_mns(cls, mns: MemoryNameSystem, wait=True) -> "ProfileReader":
        return cls(
            mns.wait_for_reader(cls.core_segment_id)
            if wait
            else mns.get_reader(cls.core_segment_id)
        )

    def query_latest(self) -> ty.Dict[MemoryID, Profile]:
        result = self.reader.query_core_segment(
            self.core_segment_id.core_segment_name, latest_snapshot=True
        )

        id_to_person = {}

        def from_aron(id: MemoryID, aron_data):
            id_to_person[id] = Profile.from_aron_ice(aron_data)

        self.reader.for_each_instance_data(from_aron, result)
        return id_to_person

    def fetch_latest_instance(
        self,
        updated_ids: ty.Optional[ty.List[MemoryID]] = None,
    ):
        if updated_ids is None:
            memory = self.reader.query_latest(self.core_segment_id)

            latest_snapshot = None

            core_seg = memory.coreSegments[self.core_segment_id.core_segment_name]
            for prov_seg in core_seg.providerSegments.values():
                for entity in prov_seg.entities.values():
                    for snapshot in entity.history.values():
                        if latest_snapshot is None:
                            latest_snapshot = snapshot
                        elif (
                            latest_snapshot.id.timestampMicroSeconds
                            < snapshot.id.timestampMicroSeconds
                        ):
                            latest_snapshot = snapshot
        else:
            for up_id in updated_ids:
                assert self.core_segment_id.contains(up_id)

            latest_snapshot_id = max(updated_ids, key=lambda i: i.timestamp_usec)
            latest_snapshot = self.reader.query_snapshot(latest_snapshot_id)

        if not latest_snapshot:
            return None

        latest_instance = latest_snapshot.instances[0]
        return latest_instance



from typing import List
from typing import Optional

import numpy as np

from armarx_memory.core import MemoryID
from armarx_memory.client import Commit

from armarx_memory.segments.base_segment import BaseClient
from armarx_memory.segments.base_segment import BaseReader
from armarx_memory.segments.base_segment import BaseWriter


class Anticipation(object):
    def __init__(self, prediction: List[str]):
        self.prediction = prediction

    def to_aron(self) -> "armarx.aron.data.dto.GenericData":
        from armarx_memory.aron.conversion import to_aron

        return to_aron({"prediction": self.prediction})

    @classmethod
    def from_aron(cls, dto: "armarx.aron.data.dto.GenericData"):
        from armarx_memory.aron.conversion import from_aron

        return cls(**from_aron(dto))


class AnticipationClientBase(BaseClient):

    core_segment_id = MemoryID("Reasoning", "Anticipation")

    default_entity_name = "prediction"


class AnticipationWriter(AnticipationClientBase, BaseWriter):
    pass


class AnticipationReader(AnticipationClientBase, BaseReader):
    pass



from typing import List, Union

from .layer import Layer


class Stage:
    def __init__(self, component: str, commit_on_exit=False, client=None):
        self.component = component
        self.commit_on_exit = commit_on_exit
        self.client = client

        self.layers: List[Layer] = []

        self._interaction_component: str = ""
        self._interaction_layers: List[str] = []

    def layer(self, name: str) -> Layer:
        """
        Create a layer.
        :param name: The layer's name.
        :return: The layer.
        """
        layer = Layer(self.component, name)
        self.layers.append(layer)
        return layer

    def add(self, layer_or_layers: Union[Layer, List[Layer]]):
        if isinstance(layer_or_layers, Layer):
            layer = layer_or_layers
            self.layers.append(layer)
        else:
            layers = list(layer_or_layers)
            self.layers += layers

    def reset(self):
        self.layers.clear()
        self._interaction_component = ""
        self._interaction_layers.clear()

    def origin_layer(self, layer_name="Origin", id="Origin", scale=1.0) -> Layer:
        from .elements.elements import Pose

        layer = self.layer(layer_name)
        layer.add(Pose(id, scale=scale))
        return layer

    def request_interaction(self, layer: Layer):
        self._interaction_component = layer.component
        self._interaction_layers.append(layer.name)

    def __repr__(self):
        return "<{} with {} layers>".format(self.__class__.__name__, len(self.layers))

    def __enter__(self):
        """
        Makes a stage usable in a with statement:
        with arviz.begin_stage() as stage:
            with stage.layer("MyLayer") as layer:
                layer.elements += [ ... ]
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.commit_on_exit and self.client is not None:
            self.client.commit(self)



from typing import Dict
from functools import lru_cache

from armarx_robots.basic_robot import Robot

from armarx import KinematicUnitInterfacePrx
from armarx import ViewSelectionInterfacePrx
from armarx import ViewTargetBase

from armarx_robots.statechart import StatechartExecutor


class A3(Robot):
    """
    ARMAR-III

    .. highlight:: python
    .. code-block:: python

        from armarx.robots import A3
        robot = A3()
        robot.say('hello world')
    """

    profile_name = "Armar3Real"

    def __init__(self):
        super().__init__()
        self.on_connect()

    @property
    @lru_cache(1)
    def kinematic_unit(self):
        return KinematicUnitInterfacePrx.get_proxy("Armar3KinematicUnit")

    def on_connect(self):
        super().on_connect()

        class GazeSelection:
            def __init__(self):
                self.view_selection = ViewSelectionInterfacePrx.get_proxy()

            def fixate(self, target, timeout=0):
                self.view_selection.addManualViewTarget(target)

        self.gaze = GazeSelection()

    def move_to(self, state_parameters=None):
        s = StatechartExecutor(self.profile_name, "PlatformGroup", "MovePlatform")
        return s.run(state_parameters, True)

    def bring_object(self, state_parameters=None):
        s = StatechartExecutor(self.profile_name, "BringObject", "")
        return s.run(state_parameters, True)

    def grasp_object(self, state_parameters=None):
        s = StatechartExecutor(
            self.profile_name, "GraspObjectGroup", "GraspSingleObject"
        )
        return s.run(state_parameters, True)



import logging
import time
import json
import os

from abc import ABC
from abc import abstractmethod

from functools import lru_cache

from armarx import EmergencyStopMasterInterfacePrx
from armarx import EmergencyStopState
from armarx import KinematicUnitInterfacePrx
from armarx import HandUnitInterfacePrx

from armarx_robots.speech import TextStateListener
from armarx_robots.statechart import StatechartExecutor
from armarx_robots.arms import Bimanual


logger = logging.getLogger(__name__)


class Robot(ABC, Bimanual):
    """
    Convenience class
    """

    def __init__(self):
        self._text_state_listener = TextStateListener()

    def on_connect(self):
        self._text_state_listener.on_connect()

        # from armarx import ElasticFusionInterfacePrx
        # self._fusion = ElasticFusionInterfacePrx.get_proxy()

    @property
    @abstractmethod
    def kinematic_unit(self):
        pass

    @property
    @lru_cache(1)
    def emergency_stop(self):
        return EmergencyStopMasterInterfacePrx.get_proxy()

    @property
    @lru_cache(1)
    def gaze(self):
        from armarx import GazeControlInterfacePrx

        return GazeControlInterfacePrx.get_proxy()

    @property
    @lru_cache(1)
    def navigator(self):
        from armarx import PlatformNavigatorInterfacePrx

        return PlatformNavigatorInterfacePrx.get_proxy()

    @property
    @abstractmethod
    def profile_name(self) -> str:
        pass

    def __str__(self) -> str:
        return f"Robot - {self.profile_name}"

    def load_robot_config(self):
        config_path = os.path.dirname(os.path.abspath(__file__))
        config_path = os.path.join(config_path, "robot_config.json")
        with open(config_path) as f:
            robot_config = json.load(f)
        return robot_config

    def what_can_you_see_now(self, state_parameters=None):
        statechart = StatechartExecutor(
            self.profile_name, "ScanLocationGroup", "WhatCanYouSeeNow"
        )
        return statechart.run(state_parameters, True)

    def handover(self, state_parameters=None):
        statechart = StatechartExecutor(
            self.profile_name, "HandOverGroup", "ReceiveFromRobot"
        )
        return statechart.run(state_parameters, True)

    def say(self, text):
        """
        Verbalizes the given text.  SSML markup is supported
        For exmaple, to verbalize in a different language use

        .. highlight:: python
        .. code-block:: python

            robot = Robot()
            robot.say('<speak><voice language="de-de">Hallo Welt</voice></speak>')

        ..see:: armarx.speech.TextStateListener.say()
        """
        self._text_state_listener.say(text)

    def scan_scene(self):
        # self._fusion.reset()
        for yaw in [-0.3, 0.3]:
            self.gaze.setYaw(yaw)
            time.sleep(0.3)
        self.gaze.setYaw(0.0)

    def stop(self):
        """
        Sets the soft emergency stop flag

        If supported by the robot then now motor commands are sent to the
        hardware
        """
        self.emergency_stop.setEmergencyStopState(
            EmergencyStopState.eEmergencyStopActive
        )



import enum

from armarx_memory.core import MemoryID
from armarx_memory.client.detail import SpecialClientBase as scb


class TextToSpeechStateWriter(scb.SpecialWriterBase):
    @classmethod
    def _get_aron_class(cls):
        return TextToSpeechState

    @classmethod
    def _get_default_core_segment_id(cls):
        return TextToSpeechState.core_segment_id


class TextToSpeechStateReader(scb.SpecialReaderBase):
    @classmethod
    def _get_aron_class(cls):
        return TextToSpeechState

    @classmethod
    def _get_default_core_segment_id(cls):
        return TextToSpeechState.core_segment_id


class TextToSpeechState:
    class Event(enum.IntEnum):
        STARTED = 0
        FINISHED = 1

    core_segment_id = MemoryID("Speech", "TextToSpeechState")
    Reader = TextToSpeechStateReader
    Writer = TextToSpeechStateWriter

    def __init__(
        self,
        event: Event,
        text: str,
        tts_snapshot_id: MemoryID,
    ):
        self.event = event
        self.text = text
        self.tts_snapshot_id = tts_snapshot_id

    def to_aron(self) -> "armarx.aron.data.dto.GenericData":
        from armarx_memory.aron.conversion import to_aron

        dto = to_aron(
            {
                "event": self.event,
                "text": self.text,
                "ttsSnapshotID": self.tts_snapshot_id,
            }
        )
        return dto

    @classmethod
    def from_aron(cls, dto: "armarx.aron.data.dto.GenericData"):
        from armarx_memory.aron.conversion import from_aron

        d = from_aron(dto)
        d["tts_snapshot_id"] = d.pop("ttsSnapshotID")
        return cls(**d)

    def __repr__(self):
        return "<{c} event='{e}' tts ID={tts} text={text}>".format(
            c=self.__class__.__name__,
            e=self.event,
            tts=self.tts_snapshot_id,
            text=self.text,
        )



"""
Module containing all the logic to handle and import slice files
"""
import os
import sys
import logging
import time

import warnings

import inspect

from importlib.abc import MetaPathFinder
import importlib
import types

import Ice


from .ice_manager import get_proxy
from .ice_manager import get_topic
from .ice_manager import register_object
from .ice_manager import wait_for_proxy

from .cmake_helper import get_include_path
from .cmake_helper import get_dependencies

from .name_helper import slice_mapping

logger = logging.getLogger(__name__)


def load_armarx_slice(armarx_package_name: str, filename: str):
    """
    ..deprecated:: add your slice file to a project's VariantInfo-*.xml instead
    """
    warnings.warn(
        "Add the slice definition to VariantInfo-*.xml instead.", DeprecationWarning
    )
    for c in sys.meta_path:
        if isinstance(c, ArmarXProxyFinder):
            c._load_armarx_slice(armarx_package_name, filename)
            c.update_loaded_modules()


class ArmarXProxyFinder(MetaPathFinder):
    """
    The ArmarXProxyFinder class

    Searches all known proxy/topic definitions as specified by
    config.get_packages() and adds them as available module to Python

    ..see:: config.get_package
    ..see:: importlib.abc.MetaPathFinder
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # all namespaces as specified by the slice definitions
        self.package_namespaces = {"armarx", "visionx"}
        # all patched interfaces
        self.patched_definitions = set()
        self.loaded_slice_files = set()
        # mapping between fullname of the proxies/topics and variant info
        self.mapping = slice_mapping
        for _, v in self.mapping.items():
            python_package_name, _type_name = v.fullname.rsplit(".", 1)
            self.package_namespaces.add(python_package_name)

    def _load_armarx_slice(self, armarx_package_name: str, filename: str):
        """
        Simple helper function to load a slice definition file.

        Loads a slice definition file from a project. Definitions in the imported
        slice file are then available through the python import function.

        :raises IOError: if the slice file was not found
        :param armarx_package_name: name of the armarx package
        :param filename: relative path to the slice interface
        """
        package_dependencies = get_dependencies(armarx_package_name)
        package_dependencies.append(armarx_package_name)

        include_paths = ["-I{}".format(Ice.getSliceDir())]

        for package_name in package_dependencies:
            interface_include_path = get_include_path(package_name)
            if interface_include_path:
                include_paths.extend(interface_include_path)
            else:
                logger.error("Include path for project %s is empty", package_name)
                raise Exception(f"Invalid include path for project {package_name}")

        filename = os.path.join(
            include_paths[-1], armarx_package_name, "interface", filename
        )
        filename = os.path.abspath(filename)

        search_path = " -I".join(include_paths)
        logger.debug("Looking for slice files in %s", search_path)
        if not os.path.exists(filename):
            raise IOError("Path not found: " + filename)
        Ice.loadSlice("{} --underscore --all {}".format(search_path, filename))

    def update_loaded_modules(self):
        """
        Update
        """
        for _, variant_info in self.mapping.items():

            if variant_info.fullname in self.patched_definitions:
                continue

            module_name, x = variant_info.fullname.rsplit(".", maxsplit=1)

            if module_name in sys.modules:
                module = sys.modules[module_name]
                if hasattr(module, x):
                    self.patch_slice_definition(variant_info)

    def find_spec(self, fullname, path, target=None):
        """
        ..see:: importlib.abc.MetaPathFinder.find_spec
        """
        if not fullname in self.mapping:
            return None

        variant_info = self.mapping.get(fullname)

        loaded_slice = f"{variant_info.package_name}/{variant_info.include_path}"
        if loaded_slice in self.loaded_slice_files:
            self.update_loaded_modules()
            return None

        self._load_armarx_slice(variant_info.package_name, variant_info.include_path)

        self.loaded_slice_files.add(loaded_slice)

        if variant_info.type == "Class":
            return None

        self.patch_slice_definition(variant_info)

        self.update_loaded_modules()

        return None

    def patch_slice_definition(self, variant_info):
        """
        Adds get_proxy, get_topic, and other methods to the imported interface
        """
        default_name = variant_info.default_name

        # already patched. No need to do it again.
        if variant_info.fullname in self.patched_definitions:
            return

        package_name, type_name = variant_info.fullname.rsplit(".", 1)

        mod = importlib.import_module(package_name)

        if not hasattr(mod, type_name):
            return

        cls = getattr(mod, type_name)

        def _get_default_topic(cls, name=None):
            return get_topic(cls, name)

        def _get_default_proxy(cls, name=None):
            return get_proxy(cls, name)

        def _wait_for_default_proxy(cls, name=None, timeout=0):
            return wait_for_proxy(cls, name, timeout)

        def _register_object(self, name: str):
            return register_object(self, name)

        if type_name.endswith("Prx"):
            proxy_class = cls
        elif hasattr(mod, type_name + "Prx"):
            proxy_class = getattr(mod, type_name + "Prx")
        else:
            return

        cls.get_topic = types.MethodType(_get_default_topic, proxy_class)
        cls.get_proxy = types.MethodType(_get_default_proxy, proxy_class)
        cls.wait_for_proxy = types.MethodType(_wait_for_default_proxy, proxy_class)
        cls.default_name = default_name

        self.patched_definitions.add(variant_info.fullname)



import numpy as np

from typing import Callable, Optional, Tuple, Union

from .core import make_grid


def make_spherical_grid(
    num: Union[int, Tuple[int, int]] = (50, 50),
    radius: Optional[float] = None,
) -> np.ndarray:
    """
    Make a grid of spherical coordinates (azim, elev) spanning
    [-pi, pi] x [-pi/2, pi/2].
    :param num:
        The number of samples along each dimension.
        (n_azim, n_elev) or (n) if n_azim == n_elev = n.
    :param radius:
        If specified, a constant radius dimension will be prepended
        to (azim, elev), i.e. each coordinate will be (radius, azim_i, elev_j).
    :return:
        If no radius is specified, a grid of spherical coordinates (azim, elev)
        of shape (n_azim, n_elev, 2).
        If a radius is specified, a grid of spherical coordinates (radius, azim, elev)
        of shape (n_azim, n_elev, 3).
    """
    try:
        num_azim, num_elev = num
    except TypeError:
        num_azim, num_elev = num, num

    azim_lim = (-np.pi, np.pi)
    elev_lim = (-np.pi / 2, np.pi / 2)
    azims = np.linspace(*azim_lim, num_azim)
    elevs = np.linspace(*elev_lim, num_elev)

    grid = make_grid(azims, elevs)

    if radius is not None:
        grid = np.insert(grid, 0, radius, axis=-1)

    return grid


def vis_sphere_mesh(
    name: str,
    num: Union[int, Tuple[int, int]] = (65, 33),
    radius=1.0,
    grid_spherical: Optional[np.ndarray] = None,
    color=(196, 196, 196, 255),
    spherical_color_fn: Optional[Callable[[np.ndarray], np.ndarray]] = None,
    cmap: Union[str, Callable] = "viridis",
) -> "armarx.arviz.Mesh":
    """
    Create a colored sphere mesh.
    :param name:
        The name of the mesh element.

    :param num:
        If `grid_spherical` is None, the number of vertices for azimuth and elevation.
    :param radius:
        If `grid_spherical` is None, the sphere's constant radius.
    :param grid_spherical:
        If specified, the grid of spherical coordinates of shape (..., 3)
        with entries (radius, azimuth, elevation).
        If not specified, it is constructed based on the arguments `num`
        and `radius`.
        This an also be used to draw shapes with non-constant radius.

    :param color:
        If `spherical_color_fn` is None, a constant color for all vertices.
    :param spherical_color_fn:
        If specified, determines the color of each vertex together with cmap.
        The function is passed the spherical coordinates of all vertices in a
        (N, 3) array and should return a scalar value y for each coordinate,
        which is passed through cmap to obtain the color of the vertex.
        The colormap is normalized to (0, y_max).
    :param cmap:
        The colormap which is used together with spherical_color_fn to
        determine vertex colors.

    :return:
        The sphere mesh.
    """
    from armarx.math import spherical
    from armarx import arviz as viz

    if grid_spherical is None:
        grid_spherical = make_spherical_grid(num=num, radius=radius)
    grid_cartesian = spherical.spherical2cartesian(grid_spherical)
    grid_cartesian = grid_cartesian.reshape((-1, 3), order="F")

    if spherical_color_fn is None:
        color = np.array(color)
        if color.shape == (3,):
            colors = np.zeros_like(grid_cartesian, dtype=int)
        elif color.shape == (4,):
            colors = np.insert(np.zeros_like(grid_cartesian, dtype=int), 0, 0, axis=-1)
        else:
            raise ValueError(
                f"Expected color to have shape (3,) or (4,), but got {color.shape}."
            )
        colors[:] = color

    else:
        from matplotlib import cm
        from matplotlib.colors import Normalize

        cmap: Callable
        if isinstance(cmap, str):
            cmap = cm.get_cmap(cmap)

        grid_spherical = grid_spherical.reshape((-1, 3), order="F")
        ys = spherical_color_fn(grid_spherical)

        colors = cmap(Normalize(vmin=0, vmax=np.max(ys))(ys))
        colors = (colors * 255).astype(int)

    mesh = viz.Mesh(name)
    mesh.vertices = grid_cartesian
    mesh.colors = colors
    mesh.faces = viz.Mesh.make_grid2d_faces(*num)

    return mesh



import numpy as np
from visionx import ImageProviderInterfacePrx


def visualize_pose(img: np.ndarray, pose: np.ndarray, K: np.ndarray) -> np.ndarray:
    """
    Draws a 6-D pose as coordinate system on a given 2D color image.

    :param img: the input image
    :param pose: the pose to visualize
    :param K: the instric camera parameter
    :returns: the input image with the visualized pose
    """
    import cv2

    axis_length = 100
    axis_thickness = 3
    R = pose[:3, :3]
    t = pose[:3, 3]
    rot_vec, _ = cv2.Rodrigues(R)

    points = np.identity(3) * 100
    points = np.vstack([points, np.zeros((1, 3))])

    points = np.float32([[100, 0, 0], [0, 100, 0], [0, 0, 100], [0, 0, 0]]).reshape(
        -1, 3
    )

    axis_points, _ = cv2.projectPoints(points, rot_vec, t, K, (0, 0, 0, 0))
    img = cv2.line(
        img,
        tuple(axis_points[3].ravel()),
        tuple(axis_points[0].ravel()),
        (255, 0, 0),
        3,
    )
    img = cv2.line(
        img,
        tuple(axis_points[3].ravel()),
        tuple(axis_points[1].ravel()),
        (0, 255, 0),
        3,
    )
    img = cv2.line(
        img,
        tuple(axis_points[3].ravel()),
        tuple(axis_points[2].ravel()),
        (0, 0, 255),
        3,
    )

    return img


def convert_armarx_to_depth(image: np.ndarray) -> np.ndarray:
    if image.dtype == np.uint8 and image.shape[-1] == 3:
        depth = image[:, :, 0].astype(np.uint16)
        depth += np.left_shift(image[:, :, 1].astype(np.uint16), 8)
        return depth
    logger.warning("Invalid image type")
    return None


def convert_depth_to_armarx(depth: np.ndarray) -> np.ndarray:
    if depth.dtype == np.uint16 and depth.shape[-1] == 1:
        image = np.zeros((*depth.shape[:-1], 3), dtype=np.uint8)
        # ..todo:: modulo
        image[:, :, 0] = depth[:, :]
        image[:, :, 1] = np.right_shift(depth[:, :], 8)
        return image
    logger.warning("Invalid image type")
    return None


def read_images(provider_name: str = None) -> np.ndarray:
    provider_name = provider_name or "OpenNIPointCloudProvider"

    image_provider = ImageProviderInterfacePrx.get_proxy(provider_name)

    if not image_provider:
        return None

    image_format = image_provider.getImageFormat()
    number_of_images = image_provider.getNumberImages()
    data_dimensions = (
        number_of_images,
        image_format.dimension.height,
        image_format.dimension.width,
        image_format.bytesPerPixel,
    )

    image_buffer, info = image_provider.getImagesAndMetaInfo()
    input_images = np.frombuffer(image_buffer, dtype=np.uint8).reshape(data_dimensions)

    return input_images, info



import logging
from abc import ABC
import time
import warnings

import numpy as np

from armarx_core.ice_manager import register_object
from armarx_core.ice_manager import get_topic

from visionx import ImageProcessorInterfacePrx
from visionx import ImageProviderInterface
from visionx import ImageFormatInfo
from visionx import ImageType
from armarx import MetaInfoSizeBase


logger = logging.getLogger(__name__)


class ImageProvider(ImageProviderInterface, ABC):
    """ """

    def __init__(
        self, name: str, num_images: int = 2, width: int = 640, height: int = 480
    ):
        super().__init__()
        self.name = name
        self.image_format = self._get_image_format(width, height)
        self.data_dimensions = (
            num_images,
            height,
            width,
            self.image_format.bytesPerPixel,
        )
        self.images = np.zeros(self.data_dimensions, dtype=np.uint8)
        image_size = self.image_format.bytesPerPixel * width * height
        self.info = MetaInfoSizeBase(image_size, image_size)

        self.image_topic = None
        self.proxy = None
        # self.register()

    def register(self):
        warnings.warn("Replaced with on_connect", DeprecationWarning)
        self.on_connect()

    def on_connect(self):
        """
        Register the image provider.
        """
        logger.debug("registering image processor %s", self.name)
        self.proxy = register_object(self, self.name)
        self.image_topic = get_topic(
            ImageProcessorInterfacePrx, f"{self.name}.ImageListener"
        )

    def update_image(self, images, time_provided=0):
        warnings.warn("Replaced with update_images", DeprecationWarning)
        self.update_images(images, time_provided)

    def update_images(self, images: np.ndarray, time_provided: int = 0):
        """
        Publish a new image

        :param images: the images to publish
        :param time_provided: time stamp of the images. If zero the current time will be used
        """
        self.images = images
        self.info.timeProvided = time_provided or int(time.time() * 1000.0 * 1000.0)
        if self.image_topic:
            self.image_topic.reportImageAvailable(self.name)
        else:
            logger.warning("not registered. call register() method")

    def _get_image_format(self, width, height):
        image_format = ImageFormatInfo()
        image_format.bytesPerPixel = 3
        image_format.dimension.width = width
        image_format.dimension.height = height
        image_format.type = ImageType.eRgb
        return image_format

    def getImageFormat(self, current=None):
        logger.debug("getImageFormat() %s", self.image_format)
        return self.image_format

    def getImagesAndMetaInfo(self, current=None):
        logger.debug("getImageFormat() %s", self.image_format)
        return memoryview(self.images), self.info

    def getImages(self, current=None):
        logger.debug("getImages()")
        return memoryview(self.images)

    def getNumberImages(self, current=None):
        return self.data_dimensions[0]

    def hasSharedMemorySupport(self, current=None):
        return False

    def shutdown(self, current=None):
        current.adapter.getCommunicator().shutdown()



from typing import List, Union, Tuple, Any

import numpy as np

from armarx import Vector3f
from armarx.viz.data import Element
from armarx.viz.data import Color
from armarx.viz.data import ColoredPoint


def vector3f_to_numpy(v: Vector3f):
    return np.array([v.e0, v.e1, v.e2])


def vector3f_from_numpy(a: np.ndarray):
    return Vector3f(*map(float, a))


def vector3fs_to_numpy(vs: List[Vector3f]):
    return np.array([]) if len(vs) == 0 else np.stack(list(map(vector3f_to_numpy, vs)))


def vector3fs_from_numpy(array: np.ndarray):
    return list(map(vector3f_from_numpy, array))


def color_to_numpy(c: Color) -> np.ndarray:
    return np.array([c.r, c.g, c.b, c.a])


def set_viz_color(c: Color, value):
    if len(value) == 3:
        c.r, c.g, c.b = map(int, value)
        c.a = int(c.a)
    else:
        c.r, c.g, c.b, c.a = map(int, value)


def to_viz_color(value) -> Color:
    c = Color()
    set_viz_color(c, value)
    return c


def colored_point_to_numpy_full(point: ColoredPoint):
    return np.array(
        [
            point.x,
            point.y,
            point.z,
            point.color.r,
            point.color.g,
            point.color.b,
            point.color.a,
        ]
    )


def colored_point_to_numpy_position(point: ColoredPoint):
    return np.array([point.x, point.y, point.z])


def colored_point_to_numpy_color(point: ColoredPoint):
    return np.array([point.color.r, point.color.g, point.color.b, point.color.a])


def colored_points_to_numpy_full(points: List[ColoredPoint]):
    if len(points) > 0:
        return np.stack(list(map(colored_point_to_numpy_full, points)))
    else:
        return np.array([])


def colored_points_to_numpy_positions(points: List[ColoredPoint]):
    return np.stack(list(map(colored_point_to_numpy_position, points)))


def colored_points_to_numpy_colors(points: List[ColoredPoint]):
    return np.stack(list(map(colored_point_to_numpy_color, points)))


def set_colored_point_from_numpy_position(cp: ColoredPoint, pos: np.ndarray):
    pos = np.array(pos)
    cp.x, cp.y, cp.z = map(float, pos)


def set_colored_point_from_numpy(cp: ColoredPoint, array: np.ndarray):
    array = np.array(array)
    if array.ndim != 1:
        raise ValueError(
            "Trying to set colored point from non-flat array with shape {}.\nPassed: {}".format(
                array.shape, array
            )
        )
    if len(array) not in [3, 6, 7]:
        raise ValueError(
            "Trying to set colored point with {} elements (expected 3, 6, or 7).\nPassed: {}".format(
                len(array), array
            )
        )

    if len(array) == 3:
        set_colored_point_from_numpy_position(cp, array)
    elif len(array) in [6, 7]:
        set_colored_point_from_numpy_position(cp, array[:3])
        set_viz_color(cp.color, array[3:])


def set_colored_points_from_numpy(cps: List[ColoredPoint], array: np.ndarray):
    array = np.array(array)

    if len(cps) < len(array):
        cps += [ColoredPoint() for _ in range(len(array) - len(cps))]
    elif len(cps) > len(array):
        del cps[len(array) :]
    assert len(cps) == len(array)

    for cp, a in zip(cps, array):
        set_colored_point_from_numpy(cp, a)


def colored_point_from_numpy(point: np.ndarray):
    cp = ColoredPoint()
    set_colored_point_from_numpy(cp, point)
    return cp


def colored_points_from_numpy(points: np.ndarray):
    return [colored_point_from_numpy(p) for p in points]



"""
Module for to interact with objects
"""

import os
import time

from typing import List

import numpy as np

from armarx.pose_helper import pose2mat
from armarx.pose_helper import mat2pose

from armarx.ice_manager import get_proxy
from armarx.ice_manager import get_topic
from armarx.slice_loader import load_armarx_slice

load_armarx_slice("RobotAPI", "objectpose/ObjectPoseProvider.ice")
load_armarx_slice("RobotAPI", "objectpose/ObjectPoseStorageInterface.ice")


from armarx.objpose import ObjectPoseStorageInterfacePrx

from armarx.objpose.data import ProvidedObjectPose
from armarx.objpose import ObjectPoseProviderPrx
from armarx.objpose import ObjectPoseTopicPrx

from armarx.data import ObjectID


def provide_poses(
    provider_name: str,
    object_names: List[str],
    poses: List[np.ndarray],
    timestamp: int = None,
):
    """
    Writes object poses to the memory

    :param provider_name: name of the provider
    :param object_name: list of object names
    :param poses: list of object poses
    """
    pose_topic = get_topic(ObjectPoseTopicPrx, "ObjectPoseTopic")
    object_poses = []
    for name, pose in zip(object_names, poses):
        pose = mat2pose(pose)
        timestamp = timestamp or (time.time() * 1e6)
        dataset, object_name = name.split("/")
        object_id = ObjectID(dataset, object_name, object_name)
        object_pose = ProvidedObjectPose(
            providerName=provider_name,
            objectPose=pose,
            objectID=object_id,
            timestampMicroSeconds=timestamp,
        )
        object_poses.append(object_pose)
    pose_topic.reportObjectPoses(provider_name, object_poses)


def get_object_poses():
    """
    Reads object poses from the robot's memory
    """
    object_memory = get_proxy(ObjectPoseStorageInterfacePrx, "ObjectMemory")
    object_poses = object_memory.getObjectPoses()
    object_poses = {o.objectID.className: o for o in object_poses}

    return {
        object_name: {
            "pose": pose2mat(object_info.objectPoseOriginal),
            "timestamp": object_info.timestamp.timeSinceEpoch.microSeconds / 1e6,
            "provider_name": object_info.providerName,
            "confidence": object_info.confidence,
            "dataset": object_info.objectID.dataset,
        }
        for object_name, object_info in object_poses.items()
    }


def get_object_path(
    database: str = "Kitchen", object_name: str = "vitalis_cereal", suffix: str = "obj"
):
    """
    Returns the file path for an object from the object database
    """
    data_path = os.environ.get(
        "h2t__PriorKnowledgeData__PATH", "./code/h2t/PriorKnowledgeData/"
    )
    if not os.path.exists(data_path):
        return None
    data_path = os.path.join(data_path, "data/PriorKnowledgeData/objects")
    data_path = os.path.join(
        data_path, database, object_name, f"{object_name}.{suffix}"
    )
    return data_path



import re


def snake_case_to_camel_case(name: str) -> str:
    # one_two_three to oneTwoThree
    # At each _, remove the _ and capitalize the following letter.
    tokens = name.split("_")
    out = tokens[0]
    out += "".join([s.capitalize() for s in tokens[1:]])

    return out


def camel_case_to_snake_case(name: str) -> str:
    # oneTwoThree to one_two_three
    regex = re.compile("[A-Z]")

    out = ""
    latest_end = 0

    for match in regex.finditer(name):
        out += (
            name[latest_end : match.start()]
            + "_"
            + name[match.start() : match.end()].lower()
        )
        latest_end = match.end()

    out += name[latest_end:]

    return out


if __name__ == "__main__":

    assert snake_case_to_camel_case("one_two_three") == "oneTwoThree"
    assert camel_case_to_snake_case("oneTwoThree") == "one_two_three"



import numpy as np
import transforms3d as tf3d

from armarx import slice_loader
from armarx_memory.ice_conv.ice_converter import IceConverter

SLICE_INCLUDE = ("RobotAPI", "core/PoseBase.ice")


class Vector3BaseConv(IceConverter):
    @classmethod
    def _import_dto(cls):
        slice_loader.load_armarx_slice(*SLICE_INCLUDE)
        import armarx

        return armarx.Vector3Base

    def _from_ice(self, dto: "armarx.Vector3Base") -> np.ndarray:
        return np.array([dto.x, dto.y, dto.z])

    def _to_ice(self, bo: np.ndarray) -> "armarx.Vector3Base":
        Vector3Base = self.get_dto()
        x, y, z = bo
        return Vector3Base(x, y, z)


class QuaternionBaseConv(IceConverter):
    @classmethod
    def _import_dto(cls):
        slice_loader.load_armarx_slice(*SLICE_INCLUDE)
        import armarx

        return armarx.QuaternionBase

    def _from_ice(self, dto: "armarx.QuaternionBase") -> np.ndarray:
        return np.array([dto.qw, dto.qx, dto.qy, dto.qz])

    def _to_ice(self, bo: np.ndarray) -> "armarx.QuaternionBase":
        import transforms3d as tf3d

        QuaternionBase = self.get_dto()

        ori = np.array(bo)
        if ori.shape == (3, 3):
            q = tf3d.quaternions.mat2quat(ori)
        else:
            q = ori
        w, x, y, z = q
        return QuaternionBase(w, x, y, z)


class PoseBaseConv(IceConverter):
    def __init__(self):
        super().__init__()
        self.vector3_conv = Vector3BaseConv()
        self.quaternion_conv = QuaternionBaseConv()

    @classmethod
    def _import_dto(cls):
        slice_loader.load_armarx_slice(*SLICE_INCLUDE)
        import armarx

        return armarx.PoseBase

    def _from_ice(self, dto: "armarx.PoseBase") -> np.ndarray:
        import transforms3d as tf3d

        pos = self.vector3_conv.from_ice(dto.position)
        ori = self.quaternion_conv.from_ice(dto.orientation)
        ori = tf3d.quaternions.quat2mat(ori)
        mat: np.ndarray = tf3d.affines.compose(T=pos, R=ori, Z=np.ones(3))
        assert mat.shape == (4, 4)
        assert np.all(mat[3, :3] == 0)
        assert mat[3, 3] == 1
        return mat

    def _to_ice(self, bo: np.ndarray) -> "armarx.PoseBase":
        PoseBase = self.get_dto()
        T, R, _, _ = tf3d.affines.decompose(bo)
        pos = self.vector3_conv.to_ice(T)
        ori = self.quaternion_conv.to_ice(R)
        return PoseBase(position=pos, orientation=ori)



import numpy as np


def rescale(value, from_lo, from_hi, to_lo, to_hi, clip=False):
    """
    Rescale a value from interval [from_lo, from_hi] to interval [to_lo, to_hi].
    :param clip: If true, the result is clipped to [to_lo, to_hi].
    :return: The rescaled value(s).
    """
    # value in [from_lo, from_hi]
    norm = (value - from_lo) / (from_hi - from_lo)  # [0, 1]
    scaled = norm * (to_hi - to_lo) + to_lo  # [to_lo, to_hi]
    if clip and to_lo > to_hi:
        to_lo, to_hi = to_hi, to_lo
    return scaled if not clip else np.clip(scaled, to_lo, to_hi)



import os
import logging
from collections import namedtuple


from lxml import etree

from .config import get_packages

from .cmake_helper import get_data_path


ArmarXProxyInfo = namedtuple(
    "ArmarXProxyInfo",
    ["package_name", "fullname", "include_path", "default_name", "type"],
)

logger = logging.getLogger(__name__)


def get_ice_default_name(cls) -> str:
    mapping = build_mapping()
    fullname = f"{cls.__module__}.{cls.__name__}"
    if mapping.get(fullname, None):
        variant_info = mapping[fullname]
        default_name = variant_info.default_name
    else:
        default_name = cls.__name__
        default_name = default_name.rstrip("Prx")
        default_name = default_name.rstrip("Interface")
    return default_name


def _load_variant_info(armarx_package_name: str):
    mapping = dict()
    data_path = get_data_path(armarx_package_name)
    if not data_path:
        logger.warning("unable to get data path for package %s", armarx_package_name)
        return mapping
    variant_path = os.path.join(
        data_path[0],
        armarx_package_name,
        "VariantInfo-{}.xml".format(armarx_package_name),
    )
    if not os.path.isfile(variant_path):
        logger.warning(
            "variant info does not exists for package %s", armarx_package_name
        )
        return mapping
    tree = etree.parse(variant_path)
    start_pos = len(armarx_package_name + "/interface/")
    for definition_type in ["Proxy", "Topic", "Class"]:
        for nav in tree.xpath("//VariantInfo/Lib/" + definition_type):
            fullname = nav.get("typeName").replace("::", ".")
            if not "." in fullname:
                fullname = f"armarx.{fullname}"
            slice_include_path = nav.get("include")[start_pos:-2] + ".ice"
            if definition_type == "Class":
                proxy_default_name = ""
            else:
                proxy_default_name = nav.get("propertyDefaultValue")

            mapping[fullname] = ArmarXProxyInfo(
                armarx_package_name,
                fullname,
                slice_include_path,
                proxy_default_name,
                definition_type,
            )
            # also make the interface available.
            # the suffix Prx is autogenerated from the slice definition
    return mapping


def build_mapping():
    global_mapping = dict()
    armarx_packages = get_packages()
    for package_name in armarx_packages.split(","):
        global_mapping.update(_load_variant_info(package_name))
    return global_mapping


slice_mapping = build_mapping()



import armarx.arviz.load_slice

from armarx.ice_conv.ice_converter import IceConverter
from armarx.math.transform import Transform


class GlobalPoseConv(IceConverter):
    @classmethod
    def _import_dto(cls):
        return armarx.viz.data.GlobalPose

    def _to_ice(self, bo: Transform, *args, **kwargs):
        dto = self.get_dto()()
        dto.x, dto.y, dto.z = bo.translation
        dto.qw, dto.qx, dto.qy, dto.qz = map(float, bo.rot_quat)
        return dto

    def _from_ice(self, dto: armarx.viz.data.GlobalPose, *args, **kwargs):
        return Transform(
            translation=(dto.x, dto.y, dto.z), rotation=(dto.qw, dto.qx, dto.qy, dto.qz)
        )



from .Reader import Reader
from .Writer import Writer
from .MemoryNameSystem import MemoryNameSystem
from .Commit import Commit, EntityUpdate
from .MemoryListener import MemoryListener
from .Commit import MemoryID



"""
.. deprecated:: 0.20.1
   Use :py:mod:`armarx_core.ice_manager` instead.
"""

import warnings
from armarx_core.ice_manager import *

warnings.warn("Use armarx_core.ice_manager instead", DeprecationWarning)



from typing import Dict
from functools import lru_cache

from armarx_robots.armar_6 import A6
from armarx_robots.basic_robot import Robot

from armarx_core import ice_manager

from armarx import HandUnitInterfacePrx
from armarx import KinematicUnitInterfacePrx


class AD(A6):
    def __init__(self):
        super().__init__()
        self.profile_name = "ArmarDEReal"
        self.both_arms_joint_names = [
            "ArmL1_Cla1",
            "ArmL2_Sho1",
            "ArmL3_Sho2",
            "ArmL4_Sho3",
            "ArmL5_Elb1",
            "ArmL6_Elb2",
            "ArmL7_Wri1",
            "ArmL8_Wri2",
            "ArmR1_Cla1",
            "ArmR2_Sho1",
            "ArmR3_Sho2",
            "ArmR4_Sho3",
            "ArmR5_Elb1",
            "ArmR6_Elb2",
            "ArmR7_Wri1",
            "ArmR8_Wri2",
        ]

        self.on_connect()

    def on_connect(self):
        pass

    # self.kinematic_observer = KinematicUnitObserverInterfacePrx.get_proxy('ArmarDEKinematicUnitObserver')

    @property
    @lru_cache(1)
    def kinematic_unit(self):
        return ice_manager.get_proxy(KinematicUnitInterfacePrx, "KinematicUnit")



import numpy as np

from typing import Dict, Optional

from armarx_core import slice_loader

from armarx_memory.ice_conv.ice_twin import IceTwin
from armarx_memory.ice_conv.RobotAPI.Box import Box
from armarx_memory.ice_conv.RobotAPI.ObjectID import ObjectID
from armarx_memory.ice_conv.RobotAPI.PoseBase import PoseBaseConv


class ObjectPose(IceTwin):
    """
    A Python twin of `armarx.objpose.data.ObjectPose`
    from `objectpose/types.ice` in RobotAPI.
    """

    JointValues = Dict[str, float]

    def __init__(
        self,
        provider_name: str = "",
        object_id: ObjectID = None,
        object_pose_robot=None,
        object_pose_global=None,
        object_pose_original=None,
        object_pose_original_frame: str = "",
        robot_pose=None,
        local_oobb: Optional[Box] = None,
    ):
        self.provider_name = provider_name
        self.object_id = ObjectID() if object_id is None else object_id

        self.object_pose_robot = (
            np.eye(4) if object_pose_robot is None else np.array(object_pose_robot)
        )
        self.object_pose_global = (
            np.eye(4) if object_pose_global is None else np.array(object_pose_global)
        )
        self.object_pose_original = (
            np.eye(4)
            if object_pose_original is None
            else np.array(object_pose_original)
        )
        self.object_pose_original_frame = object_pose_original_frame

        self.object_joint_values: ObjectPose.JointValues = {}

        self.robot_pose = np.eye(4) if robot_pose is None else np.array(robot_pose)
        self.robot_config: ObjectPose.JointValues = {}

        # attachment

        self.confidence: float = 0.0
        self.timestamp_usec: int = -1

        self.local_oobb = local_oobb

        self.__pose_conv: Optional[PoseBaseConv] = None

    def get_aabb_robot(self):
        return self.local_oobb.to_aabb(self.object_pose_robot)

    def viz_oobb_robot(self, id: str, size_factor=1.0, **kwargs) -> "armarx.arviz.Box":
        import armarx.arviz as viz

        return viz.Box(
            id,
            pose=self.object_pose_robot @ self.local_oobb.pose,
            size=self.local_oobb.extents * size_factor,
            **kwargs,
        )

    @classmethod
    def _get_ice_cls(cls):
        slice_loader.load_armarx_slice("RobotAPI", "objectpose/object_pose_types.ice")
        from armarx.objpose.data import ObjectPose

        return ObjectPose

    def _set_from_ice(self, dto: "armarx.objpose.data.ObjectPose"):
        self.provider_name = dto.providerName
        self.object_id = ObjectID(dto.objectID)
        self.object_pose_robot = self._pose_conv.from_ice(dto.objectPoseRobot)
        self.object_pose_global = self._pose_conv.from_ice(dto.objectPoseGlobal)
        self.object_pose_original = self._pose_conv.from_ice(dto.objectPoseOriginal)
        self.object_pose_original_frame = dto.objectPoseOriginalFrame

        self.object_joint_values = dto.objectJointValues

        self.robot_config = dto.robotConfig
        self.robot_pose = self._pose_conv.from_ice(dto.robotPose)

        self.confidence = dto.confidence
        self.timestamp_usec = dto.timestamp.timeSinceEpoch.microSeconds

        if all(
            [dto.localOOBB.position, dto.localOOBB.orientation, dto.localOOBB.extents]
        ):
            self.local_oobb = Box.from_ice(dto.localOOBB)
        else:
            self.local_oobb = None

    def _set_to_ice(self, dto):
        dto.providerName = self.provider_name
        dto.objectID = self.object_id.to_ice()
        dto.objectPoseRobot = self._pose_conv.to_ice(self.object_pose_robot)
        dto.objectPoseGlobal = self._pose_conv.to_ice(self.object_pose_global)
        dto.objectPoseOriginal = self._pose_conv.to_ice(self.object_pose_original)
        dto.objectPoseOriginalFrame = self.object_pose_original_frame

        dto.objectJointValues = self.object_joint_values

        dto.robotConfig = self.robot_config
        dto.robotPose = self._pose_conv.to_ice(self.robot_pose)

        dto.confidence = self.confidence
        dto.timestamp.timeSinceEpoch.microSeconds = self.timestamp_usec

        if dto.localOOBB is not None and all(
            [dto.localOOBB.position, dto.localOOBB.orientation, dto.localOOBB.extents]
        ):
            dto.localOOBB = self.local_oobb.to_ice()
        else:
            dto.localOOBB = None

    @property
    def _pose_conv(self) -> PoseBaseConv:
        if self.__pose_conv is None:
            self.__pose_conv = PoseBaseConv()
        return self.__pose_conv

    def __repr__(self):
        return f"<{self.__class__.__name__} of {str(self.object_id)}>"



from functools import lru_cache

from visionx import ImageProviderInterfacePrx
from visionx import ReferenceFrameInterfacePrx

from armarx_vision.camera_utils import get_stereo_calibration
from armarx_vision.camera_utils import get_calibration
from armarx_vision.image_utils import read_images


class Camera:
    def __init__(self, name):
        self.provider_name = name
        self.images = property(lambda self: read_images(name))

    @property
    @lru_cache(1)
    def info(self):
        proxy = ImageProviderInterfacePrx.get_proxy(self.provider_name)
        image_format = proxy.getImageFormat()
        image_format = proxy.getImageFormat()
        width = image_format.dimension.width
        height = image_format.dimension.height

        proxy = ImageProviderInterfacePrx.get_proxy(self.provider_name)
        number_of_images = proxy.getNumberImages()

        return {
            "proxy": proxy,
            "width": width,
            "height": height,
            "num_images": number_of_images,
        }

    @property
    @lru_cache(1)
    def reference_frame_name(self):
        proxy = ReferenceFrameInterfacePrx.get_proxy(self.provider_name)
        return proxy.getReferenceFrame()

    @property
    @lru_cache(1)
    def num_images(self):
        proxy = ImageProviderInterfacePrx.get_proxy(self.provider_name)
        return proxy.getNumberImages()

    @property
    @lru_cache(1)
    def calibration(self):
        if self.num_images == 2:
            return get_stereo_calibration(self.provider_name)
        else:
            return get_calibration(self.provider_name)



"""
This module automatically injects available slice definitions into the armarx namespace
"""

import sys

from armarx_core.config import CodeGenerationType
from armarx_core.config import get_code_generation_type
from armarx_core.config import get_python_build_dir

from armarx_core.slice_loader import ArmarXProxyFinder


def inject() -> bool:
    """
    :returns: true if the slice loading is activated, false if it was already
    activated
    """
    if get_code_generation_type() == CodeGenerationType.STATIC:
        python_build_dir = get_python_build_dir()
        if python_build_dir in sys.path:
            return False
        sys.path.append(python_build_dir)
        return True
    else:
        if any(isinstance(c, ArmarXProxyFinder) for c in sys.meta_path):
            return False
        sys.meta_path.insert(0, ArmarXProxyFinder())
        return True


# auto inject
inject()



import typing as ty

from armarx_core import slice_loader

slice_loader.load_armarx_slice("RobotAPI", "armem/query.ice")

from armarx import armem
from armarx.armem import data as dto  # The ice type namespace.

from armarx_memory.core import MemoryID


class Reader:

    ReadingMemoryServerPrx = "armem.server.ReadingMemoryInterfacePrx"
    qd = armem.query.data

    def __init__(
        self,
        server: ty.Optional[ReadingMemoryServerPrx],
    ):

        self.server = server

    def query(
        self,
        queries: ty.List[armem.query.data.MemoryQuery],
    ) -> armem.data.Memory:
        """
        Perform a memory query. Return the result if successful,
        otherwise raise an exception.
        :param queries: The query(s)
        :return: The result, if successful.
        """

        inp = self.qd.Input(memoryQueries=queries, withData=True)
        result = self.server.query(inp)
        if not result.success:
            raise RuntimeError(f"Memory query failed. Reason:\n{result.errorMessage}")
        else:
            return result.memory

    def query_snapshots(
        self,
        ids: ty.List[MemoryID],
    ) -> ty.Dict[MemoryID, armem.data.EntitySnapshot]:
        """
        Query snapshots corresponding to ty.List of memory IDs.

        Each ID can refer to an entity, a snapshot or an instance. When not
        referring to an entity snapshot, the latest snapshot will be queried.

        All memory IDs must refer to the memory this reader is reading from.
        If an ID refers to another memory, the query will not find it and it
        will not be part of the result.

        :param ids: The entity, snapshot or instance IDs.
        :return: The query result, if successful.
        """

        qs_memory = []
        for snapshot_id in ids:
            if snapshot_id.timestamp_usec >= 0:
                ice_id = snapshot_id.to_ice()
                q_entity = self.qd.entity.Single(timestamp=ice_id.timestamp)
            else:
                q_entity = self.qd.entity.Single()  # Latest

            q_prov = self.qd.provider.Single(
                entityName=snapshot_id.entity_name, entityQueries=[q_entity]
            )
            q_core = self.qd.core.Single(
                providerSegmentName=snapshot_id.provider_segment_name,
                providerSegmentQueries=[q_prov],
            )
            q_memory = self.qd.memory.Single(
                coreSegmentName=snapshot_id.core_segment_name,
                coreSegmentQueries=[q_core],
            )
            qs_memory.append(q_memory)

        memory = self.query(qs_memory)

        snapshots = dict()
        for snapshot_id in ids:
            entity = (
                memory.coreSegments[snapshot_id.core_segment_name]
                .providerSegments[snapshot_id.provider_segment_name]
                .entities[snapshot_id.entity_name]
            )

            # TODO: Quick and dirty fix: Iterate over all history entries
            # TODO: And fix that if timestamp is -2 you want to have the secondlast element
            if snapshot_id.timestamp_usec >= 0:
                for datetime, entry in entity.history.items():
                    if (
                        snapshot_id.timestamp_usec
                        == datetime.timeSinceEpoch.microSeconds
                    ):
                        snapshots[snapshot_id] = entry
                        break
            else:
                snapshots[snapshot_id] = entity.history[
                    sorted(
                        entity.history.keys(),
                        key=lambda x: x.timeSinceEpoch.microSeconds,
                    )[-1]
                ]

            # snapshots[snapshot_id] = entity.history[
            #    ice_id.timestamp
            #    if snapshot_id.timestamp_usec >= 0
            #    else max()
            # ]

        return snapshots

    def query_snapshot(
        self,
        snapshot_id: MemoryID,
    ) -> armem.data.EntitySnapshot:

        return self.query_snapshots([snapshot_id])[snapshot_id]

    def query_all(
        self,
    ) -> armem.data.Memory:

        q_entity = self.qd.entity.All()
        q_prov = self.qd.provider.All(entityQueries=[q_entity])
        q_core = self.qd.core.All(providerSegmentQueries=[q_prov])
        q_memory = self.qd.memory.All(coreSegmentQueries=[q_core])
        memory = self.query([q_memory])
        return memory

    def query_core_segment(
        self,
        name: str,
        regex=False,
        latest_snapshot=False,
    ) -> armem.data.Memory:

        if latest_snapshot:
            q_entity = self.qd.entity.Single()  # Latest
        else:
            q_entity = self.qd.entity.All()
        q_prov = self.qd.provider.All(entityQueries=[q_entity])
        q_core = self.qd.core.All(providerSegmentQueries=[q_prov])
        if regex:
            q_memory = self.qd.memory.Regex(
                coreSegmentNameRegex=name, coreSegmentQueries=[q_core]
            )
        else:
            q_memory = self.qd.memory.Single(
                coreSegmentName=name, coreSegmentQueries=[q_core]
            )
        memory = self.query([q_memory])
        return memory

    def query_latest(
        self,
        memory_id: ty.Optional[MemoryID] = None,
    ) -> armem.data.Memory:
        if memory_id is None:
            memory_id = MemoryID()

        q_entity = self.qd.entity.Single()  # Latest
        if memory_id.entity_name:
            q_prov = self.qd.provider.Single(
                entityName=memory_id.entity_name, entityQueries=[q_entity]
            )
        else:
            q_prov = self.qd.provider.All(entityQueries=[q_entity])

        if memory_id.provider_segment_name:
            q_core = self.qd.core.Single(
                providerSegmentName=memory_id.provider_segment_name,
                providerSegmentQueries=[q_prov],
            )
        else:
            q_core = self.qd.core.All(providerSegmentQueries=[q_prov])

        if memory_id.core_segment_name:
            q_memory = self.qd.memory.Single(
                coreSegmentName=memory_id.core_segment_name, coreSegmentQueries=[q_core]
            )
        else:
            q_memory = self.qd.memory.All(coreSegmentQueries=[q_core])

        memory = self.query([q_memory])
        return memory

    @classmethod
    def for_each_instance_data(
        cls,
        fn: ty.Callable[[MemoryID, ty.Dict[str, ty.Any]], ty.Any],
        data: ty.Union[
            dto.Memory,
            dto.CoreSegment,
            dto.ProviderSegment,
            dto.Entity,
            dto.EntitySnapshot,
            dto.EntityInstance,
        ],
        discard_none=False,
    ) -> ty.List[ty.Any]:
        """
        Call `fn` on the data of each entity instance in `data`.

        Iterate over a memory data structure and fall `fn` on the data of
        each entity instance. The data is converted to python data structures
        beforehand.

        Example:

        def process_instance_data(id: MemoryID, data: Dict):
            print(id)
            return id

        memory = reader.query(...)
        ids = reader.for_each_instances_data(process_instance_data, memory)

        :param fn: The function to call on each instance data.
        :param data: The data structure (e.g. the result of a query).
        :param discard_none: If true, None return values are excluded from the result list.
        :return: The values returned by the calls to `fn`.
        """
        from armarx_memory.aron.conversion import from_aron

        if isinstance(data, dto.EntityInstance):
            pythonic_data: ty.Dict[str, ty.Any] = from_aron(data.data)
            memory_id = MemoryID.from_ice(data.id)
            return [fn(memory_id, pythonic_data)]

        elif isinstance(data, dto.EntitySnapshot):
            children = data.instances
        elif isinstance(data, dto.Entity):
            children = data.history.values()
        elif isinstance(data, dto.ProviderSegment):
            children = data.entities.values()
        elif isinstance(data, dto.CoreSegment):
            children = data.providerSegments.values()
        elif isinstance(data, dto.Memory):
            children = data.coreSegments.values()
        else:
            raise TypeError(f"Unexpected data of type {type(data)}: {data}")

        results = []
        for child in children:
            results += cls.for_each_instance_data(fn, child)

        if discard_none:
            results = list(filter(lambda r: r is not None, results))

        return results

    def __bool__(self):
        return bool(self.server)



"""
A module with statechart related code
"""
import logging

from typing import Dict
from typing import Any

from armarx import SimpleStatechartExecutorInterfacePrx

from armarx import StateParameterIceBase

from armarx import SingleVariantBase
from armarx import SingleTypeVariantListBase
from armarx import StringValueMapBase
from armarx import VariantBase
from armarx import ContainerType

from armarx_core.variants import hash_type_name
from armarx_core.variants import convert_to_variant_data


logger = logging.getLogger(__name__)


class StatechartExecutor(object):
    """
    A convenience class to run statecharts using the SimpleStatechartExecutor
    component
    """

    def __init__(self, profile_name: str, group_name: str, state_name: str):
        self.fullname = f"{profile_name}{group_name}"
        self.state_name = state_name
        self.executor = SimpleStatechartExecutorInterfacePrx.get_proxy()
        self.logger = logging.getLogger(self.__class__.__name__)

    #  -> Ice.Future:
    def run(
        self,
        state_parameters: Dict[str, Any] = None,
        stop_current_statechart: bool = False,
    ):
        """
        Runs the statechart via the SimpleStatechartExecutor component

        :param state_parameters: the statechart parameters
        :param stop_current_statechart:  stop an already running statechart
        :returns: a future of the execution status
        """
        self.logger.info("running state %s of group %s", self.state_name, self.fullname)

        state_parameters = state_parameters or {}
        self.logger.debug("with parameters %s", state_parameters)

        # ..todo:: check if ice_staticId is available otherwise convert the data.
        preload_libraries = []
        preload_libraries.append("::armarx::StringVariantData")
        preload_libraries.append("::armarx::DoubleVariantData")
        preload_libraries.append("::armarx::FloatVariantData")
        preload_libraries.append("::armarx::PoseBase")

        state_parameters: Dict[str, StateParameterIceBase] = {
            k: self._get_state_parameter(v) for k, v in state_parameters.items()
        }

        # preload_libraries = [v.ice_staticId() for k, v in state_parameters.items() if not isinstance(v, list)]

        self.logger.debug("preloading libraries %s", preload_libraries)
        self.executor.preloadLibrariesFromHumanNames(preload_libraries)

        if not self.executor.hasExecutionFinished():
            self.logger.warning("another statechart is currently executed.")
            if stop_current_statechart:
                self.logger.debug("stopping running statechart")
                self.executor.stopImmediatly()
            else:
                self.logger.debug("waiting for running statechart")
                return self.executor.waitUntilStatechartExecutionIsFinishedAsync()
        remote_name = f"{self.fullname}RemoteStateOfferer"
        self.logger.debug(
            "running state %s of remote state %s", self.state_name, remote_name
        )
        if not self.executor.startStatechart(
            remote_name, self.state_name, state_parameters
        ):
            self.logger.warning(
                "unable to run state %s of state %s", self.state_name, remote_name
            )
        return self.executor.waitUntilStatechartExecutionIsFinishedAsync()

    def get_output_parameters(self):
        """
        Returns the output parameters of the statechart
        """
        self.logger.debug("getting statechart output parameters")
        return self.executor.getOutputParameters()

    def stop(self):
        """
        Stops the current running statechart
        """
        self.logger.debug("stopping statechart execution")
        self.executor.stopImmediatly()

    @staticmethod
    def _get_state_parameter(data):
        converted_parameter = StatechartExecutor._convert_parameter(data)
        return StateParameterIceBase(value=converted_parameter, set=True)

    @staticmethod
    def _convert_parameter(data):
        """ """

        def _convert(v):
            v = convert_to_variant_data(v)
            v = StatechartExecutor._convert_parameter(v)
            return v

        if isinstance(data, (list, tuple)):
            logger.warning("not tested yet")
            container_ice_id = SingleTypeVariantListBase.ice_staticId()
            # ..todo:: do we need a variant base?
            elements = [_convert(i) for i in data]
            parameter_ice_id = elements[0]._typeContainer.typeId
            type_container = ContainerType(
                ContainerType(None, parameter_ice_id), container_ice_id
            )
            return SingleTypeVariantListBase(type_container, elements)
        elif isinstance(data, dict):
            logger.warning("not tested yet")
            container_ice_id = StringValueMapBase.ice_staticId()
            elements = {k: _convert(v) for k, v in data}
            parameter_ice_id = list(elements.values())[0]._typeContainer.typeId
            type_container = ContainerType(
                ContainerType(None, parameter_ice_id), container_ice_id
            )
            return StringValueMapBase(type_container, elements)
        else:
            data = convert_to_variant_data(data)
            parameter_ice_id = data.ice_staticId()
            type_container = ContainerType(None, parameter_ice_id)
            variant = VariantBase(data, hash_type_name(parameter_ice_id))
            return SingleVariantBase(type_container, variant)

    def __str__(self):
        return f"{self.__class__.__name__}{self.fullname}"



import os

from armarx_core import slice_loader

from armarx_memory.ice_conv.ice_twin import IceTwin


class ObjectID(IceTwin):
    """
    A Python twin of the Ice DTO class `armarx.data.ObjectID`
    from `ArmarXObjects/ArmarXObjectsTypes.ice` in RobotAPI.
    """

    DEFAULT_PRIOR_KNOWLEDGE_PACKAGE = "PriorKnowledgeData"
    SEP = "/"

    def __init__(self, dataset="", class_name="", instance_name=""):
        self.dataset = dataset
        self.class_name = class_name
        self.instance_name = instance_name

    @classmethod
    def from_str(cls, id: str) -> "ObjectID":
        return cls(*id.split(cls.SEP))

    def get_dir(self, package=DEFAULT_PRIOR_KNOWLEDGE_PACKAGE):
        return os.path.join(package, "objects", self.dataset, self.class_name)

    def get_filename(self, extension: str, suffix=""):
        if not extension.startswith("."):
            extension = "." + extension
        return f"{self.class_name}{suffix}{extension}"

    def get_filepath(
        self,
        extension: str,
        suffix="",
        package=DEFAULT_PRIOR_KNOWLEDGE_PACKAGE,
        resolve=False,
    ):
        filepath = os.path.join(
            self.get_dir(package=package),
            self.get_filename(extension=extension, suffix=suffix),
        )
        return self.resolve_filepath(filepath) if resolve else filepath

    def get_viz_kwargs(self, package=DEFAULT_PRIOR_KNOWLEDGE_PACKAGE, use_wrl=False):
        return dict(
            project=package, filename=self.get_filepath(".wrl" if use_wrl else ".xml")
        )

    @classmethod
    def resolve_filepath(cls, filepath: str):
        from armarx import cmake_helper

        package_name = filepath.split(os.path.sep, maxsplit=1)[0]
        package_data_dir = cmake_helper.get_data_path(package_name)[0]
        return os.path.join(package_data_dir, filepath)

    @classmethod
    def _get_ice_cls(cls):
        slice_loader.load_armarx_slice(
            "RobotAPI", "ArmarXObjects/ArmarXObjectsTypes.ice"
        )
        from armarx.data import ObjectID

        return ObjectID

    def _set_from_ice(self, dto: "armarx.data.ObjectID"):
        self.dataset = dto.dataset
        self.class_name = dto.className
        self.instance_name = dto.instanceName

    def _set_to_ice(self, dto: "armarx.data.ObjectID"):
        dto.dataset = self.dataset
        dto.className = self.class_name
        dto.instanceName = self.instance_name

    def __repr__(self) -> str:
        return "<{} dataset='{}' class_name='{}' instance_name='{}'>".format(
            self.__class__.__name__, self.dataset, self.class_name, self.instance_name
        )

    def __str__(self) -> str:
        items = [self.dataset, self.class_name]
        if self.instance_name:
            items.append(self.instance_name)
        return self.SEP.join(items)



import abc

import numpy as np

from typing import List, Tuple, Union, Callable, Any, Optional

from armarx.math import rescale


class SizeModel(abc.ABC):
    class Context:
        def __init__(self, i=0, x=0.0, y=0.0):
            self.i = i
            self.x = x
            self.y = y

    def __init__(self):
        pass

    @abc.abstractmethod
    def get_size(
        self,
        voxel_size: np.ndarray,
        c: Context,
    ) -> Optional[np.ndarray]:
        pass

    def set_value_limits(
        self,
        vmin: float,
        vmax: float,
    ):
        pass


class FixedScaleSizeModel(SizeModel):
    def __init__(self, scale=1.0):
        super().__init__()
        self.scale = scale

    def get_size(self, voxel_size, context):
        return self.scale * voxel_size


class ValueProportionalSizeModel(SizeModel):
    """
    Size model that scales voxel proportionally to their value.
    """

    def __init__(
        self,
        from_lo_t=0.0,
        from_hi_t=1.0,
        to_lo_scale=0.0,
        to_hi_scale=1.0,
        fixed_scale=1.0,
        value_limits=(0.0, 1.0),
        min_absolute_size=1e-4,
    ):
        super().__init__()

        self.from_lo_t = from_lo_t
        self.from_hi_t = from_hi_t
        self.to_lo_scale = to_lo_scale
        self.to_hi_scale = to_hi_scale
        self.fixed_scale = fixed_scale

        self.from_lo = 0.0
        self.from_hi = 1.0
        self.set_value_limits(value_limits[0], value_limits[1])

        self.min_absolute_size = min_absolute_size

    def set_value_limits(self, vmin: float, vmax: float):
        self.from_lo = rescale(self.from_lo_t, 0.0, 1.0, vmin, vmax)
        self.from_hi = rescale(self.from_hi_t, 0.0, 1.0, vmin, vmax)

    def get_size(self, voxel_size, context):
        prop = rescale(
            context.y,
            from_lo=self.from_lo,
            from_hi=self.from_hi,
            to_lo=self.to_lo_scale,
            to_hi=self.to_hi_scale,
            clip=True,
        )
        size = prop * self.fixed_scale * voxel_size
        return size if min(size) > self.min_absolute_size else np.zeros_like(size)



from .widgets import (
    Label,
    LineEdit,
    ComboBox,
    IntSpinBox,
    IntSlider,
    FloatSpinBox,
    FloatSlider,
    Button,
    ToggleButton,
    CheckBox,
    HBoxLayout,
    VBoxLayout,
    GridLayout,
    GroupBox,
    VSpacer,
    HSpacer,
)
from .widgets.ndarray import NdArrayWidget

from .client import Tab, Client



from typing import Union, List, Dict, Any


class IceTwin:
    """
    A python class representing a python-twin of an class defined in Ice.

    Hint: Kind of deprecated. Use IceConverter from armarx.ice_conv.ice_converter instead.

    Usage:

    class MyClass(IceTwin):  # Derive from IceTwin

        def __init__(self, data: int):
            self.data = data


        # Override _protected functions

        @classmethod
        def _get_ice_cls(cls):
            return armarx.dto.MyClass

        def _set_from_ice(self, dto: armarx.dto.MyClass):
            self.data = dto.data

        def _set_to_ice(self, dto: armarx.dto.MyClass):
            dto.data = self.data
    """

    @classmethod
    def from_ice(cls, ice):
        if isinstance(ice, list):
            return list(map(cls.from_ice, ice))
        elif isinstance(ice, dict):
            return {k: to_ice(v) for k, v in ice.items()}
        else:
            self = cls()
            self._set_from_ice(ice)
            return self

    def to_ice(self):
        cls = self._get_ice_cls()
        ice = cls()
        self._set_to_ice(ice)
        return ice

    def set_from_ice(self, dto):
        self._set_from_ice(dto)

    def set_to_ice(self, dto):
        self._set_to_ice(dto)

    @classmethod
    def _get_ice_cls(cls):
        """Return the Ice DTO class."""
        raise NotImplementedError(
            f"{cls.__name__} does not implement `_get_ice_cls()`."
        )

    def _set_from_ice(self, dto):
        """Set `self` from `dto`."""
        raise NotImplementedError(
            f"{self.__class__.__name__} cannot be set from ice. "
            "\nImplement `_set_from_ice()` to allow it."
        )

    def _set_to_ice(self, dto):
        """Set `dto` from `self`."""
        raise NotImplementedError(
            f"{self.__class__.__name__} cannot set to ice. "
            "\nImplement `_set_to_ice()` to allow it.`"
        )


def to_ice(twin: Union[IceTwin, List, Dict]):
    if isinstance(twin, IceTwin):
        return twin.to_ice()
    elif isinstance(twin, list):
        return [to_ice(t) for t in twin]
    elif isinstance(twin, dict):
        return {k: to_ice(t) for k, t in twin.items()}


def __swap_assignments(code: str):
    """
    Helps you write _set_to_ice() after writing _set_from_ice().
    """
    lines = code.splitlines(keepends=False)
    out_lines = []
    for line in lines:
        sides = [s.strip() for s in line.split("=")]
        if len(sides) == 2:
            out_lines.append("{} = {}".format(sides[1], sides[0]))
        else:
            out_lines.append(line)

    print("".join(["{}\n".format(l) for l in out_lines]))



import logging

import dataclasses as dc
import typing as ty

from .options import ConversionOptions


def dataclass_to_dict(
    obj,
    logger: ty.Optional[logging.Logger] = None,
) -> ty.Dict[str, ty.Any]:
    """
    Deeply converts a dataclass to a dict.

    :param obj: An object of a dataclass.
    :param logger: An optional logger.
    :return: A dict containing pythonic data types.
    """
    return dc.asdict(obj)


def dataclass_from_dict(
    cls,
    data: ty.Dict,
    logger: ty.Optional[logging.Logger] = None,
):
    """
    Deeply converts a dictionary with pythonic data types
    to an instance of the given dataclass.

    :param cls: The dataclass.
    :param data: The data.
    :param logger: A logger.
    :return: An instance of the dataclass.
    """

    if logger is not None:
        logger.info(f"\nConstruct class {cls.__name__} from a {type(data)} ...")

    try:
        field_types = cls.__annotations__
    except AttributeError:
        # Not a dataclass. Can just try to deliver kwargs. Or return data.
        try:
            return cls(**data)
        except TypeError:
            return data

    # Build kwargs for cls.
    kwargs = dict()
    for field_name, value in data.items():
        field_type = field_types[field_name]

        try:
            origin = field_type.__origin__
        except AttributeError:
            origin = None

        if logger is not None:
            value_type = type(value)
            logger.debug(
                f"- Field '{field_name}' of type: {field_type}"
                f"\n- origin: {origin}"
                f"\n- data type: {value_type}"
            )

        if origin in (ty.List, list):
            [vt] = field_type.__args__
            if logger is not None:
                logger.debug(f"> Process list of {vt} ")
            field_value = [dataclass_from_dict(vt, v) for v in value]

        elif origin in (ty.Dict, dict):
            kt, vt = field_type.__args__
            if logger is not None:
                logger.debug(f"> Process dict {kt} -> {vt}")
            if value is not None:
                field_value = {
                    kt(k): dataclass_from_dict(vt, v) for k, v in value.items()
                }
            else:
                field_value = None

        else:
            if logger is not None:
                logger.debug(f"> Process other: {field_type}")
            try:
                field_value = dataclass_from_dict(field_type, value)
            except AttributeError:
                # Cannot convert.
                field_value = value

        kwargs[field_name] = field_value

    return cls(**kwargs)
